""" Class to save the current request object """


class RequestHandler:
    """ Class to save the current request object """

    def __init__(self):
        self.request = None

    def set_request(self, request):
        """ Set the request object """
        self.request = request

    def get_request(self):
        """ Get the request object """
        return self.request

# Usage:


# Create a single instance of RequestHandler to be used throughout a module
# request_handler = RequestHandler()


# def set_local_request(request):
#     request_handler.set_request(request)



"""
Framework abstraction layer
"""
import os
import importlib

DEFAULT_FRAMEWORK = ''

def get_current_framework():
    """
    Get the current framework
    """
    return os.environ.get('CURRENT_FRAMEWORK', DEFAULT_FRAMEWORK)


framework = os.environ.get('CURRENT_FRAMEWORK', DEFAULT_FRAMEWORK).lower()
if framework not in ['chalice', 'fastapi', 'flask']:
    raise ValueError("Unsupported or undefined CURRENT_FRAMEWORK:" +
                     f" '{framework}'")

module_base_path = f"genericsuite.{framework}lib"
module_path = f"{module_base_path}.framework_abstraction"

try:
    framework_module = importlib.import_module(module_path)
    if not framework_module or not framework_module.FRAMEWORK_LOADED:
        raise ImportError(f"Framework '{module_path}' could no be loaded" +
            " [FAL-E010]" +
            f" | FRAMEWORK_LOADED={framework_module.FRAMEWORK_LOADED}")
except ImportError as err:
    raise ImportError(f"Unable to import module [1]: {err}") from None

try:
    create_app_path = f'{module_base_path}.util.create_app'
    module_create_app = importlib.import_module(create_app_path)
except ImportError as err:
    raise ImportError(f"Unable to import module [2]: {err}") from None


class FrameworkClass(framework_module.FrameworkClass):
    """
    Framkework class cloned from the selected framework super class.
    E.g. chalice, fastapi, flask:
        from flask import Flask
        from fastapi import FastAPI
        from chalice import Chalice
    Will be:
        from genericsuite.util.framework_abs_layer import FrameworkClass as Flask
        from genericsuite.util.framework_abs_layer import FrameworkClass as FastAPI
        from genericsuite.util.framework_abs_layer import FrameworkClass as Chalice
    This class is the one to be imported by the project modules.
    """


class Request(framework_module.Request):
    """
    Request class cloned from the selected Request framework super class.
    This class is the one to be imported by the project modules.
    """


class Response(framework_module.Response):
    """
    Response class cloned from the selected Response framework super class.
    This class is the one to be imported by the project modules.
    """


class Blueprint(framework_module.Blueprint):
    """
    Blueprint class cloned from the selected Blueprint framework super class.
    This class is the one to be imported by the project modules.
    """


class BlueprintOne(framework_module.BlueprintOne):
    """
    Blueprint wrapper to add authorization, other data and schema validation
    to requests.
    """



"""
Generic Endpoint builder for FastAPI
"""
from typing import Optional, Union

from pprint import pprint

from fastapi import FastAPI, Query, Body, Depends
from fastapi import Request as FaRequest
from pydantic import BaseModel

# from genericsuite.util.framework_abs_layer import FrameworkClass as FastAPI
# from genericsuite.util.framework_abs_layer import Response
from genericsuite.fastapilib.framework_abstraction import Response
from genericsuite.fastapilib.util.blueprint_one import BlueprintOne
from genericsuite.fastapilib.util.dependencies import (
    get_current_user,
    build_request,
)

from genericsuite.util.config_dbdef_helpers import get_json_def
from genericsuite.util.generic_endpoint_helpers import GenericEndpointHelper
from genericsuite.util.app_logger import log_debug
from genericsuite.util.utilities import return_resultset_jsonified_or_exception
from genericsuite.config.config_from_db import app_context_and_set_env

from genericsuite.config.config import Config

DEBUG = False

class Endpoint(BaseModel):
    """
    A model for endpoint definitions.
    """
    name: str
    route: str
    method: str
    response_model: Optional[str] = None


def generate_blueprints_from_json(
    app: FastAPI,
    json_file: str = "endpoints",
) -> None:
    """
    Generates routes from a JSON file and registers them with
    the FastAPI app.

    Args:
        app (Chalice): The Chalice app to register the blueprints with.
        json_file (str, optional): The JSON file containing blueprint
        definitions. Defaults to "endpoints".
    """
    settings = Config()
    cnf_db_base_path = settings.GIT_SUBMODULE_LOCAL_PATH
    definitions = get_json_def(json_file, f'{cnf_db_base_path}/backend', [])
    for definition in definitions:
        bp_name = definition['name']
        url_prefix = f"/{definition.get('url_prefix', bp_name)}"

        if DEBUG:
            log_debug(
                'GENERATE_BLUEPRINTS_FROM_JSON |' +
                f' bp_name: {bp_name}' +
                f' url_prefix: {url_prefix}'
            )

        # Add routes to the router
        for route in definition['routes']:
            other_params = {
                "name": bp_name
            }
            other_params["params"] = route['params']

            if route['handler_type'] == "GenericEndpointHelper":
                route_handler = generic_route_handler
            else:
                route_handler = route['view_function']

            if DEBUG:
                log_debug(
                    'GENERATE_BLUEPRINTS_FROM_JSON |' +
                    f" route_endpoint: {route['endpoint']}" +
                    f" route_methods: {route['methods']}" +
                    f" route_handler_type: {route['handler_type']}" +
                    f" route_handler: {route_handler}" +
                    f" other_params: {other_params}"
                )

            for method in route['methods']:
                endpoint = {
                    "name": bp_name,
                    "route": f"{url_prefix}/{route['endpoint']}"
                        if route['endpoint'] and route['endpoint'] != '/'
                        else url_prefix,
                    "method": method,
                    "response_model": route.get("response_model")
                }
                endpoint_obj = Endpoint(**endpoint)

                # Assuming the response model is a Pydantic model and already imported
                response_model = globals()[endpoint_obj.response_model] \
                    if "response_model" in route else None

                other_params["method"] = endpoint_obj.method.lower()
                if other_params["method"] == 'get':
                    app.get(endpoint_obj.route, response_model=response_model,
                            tags=[bp_name])(
                        create_endpoint_function(other_params)
                    )
                elif other_params["method"] == 'post':
                    app.post(endpoint_obj.route, response_model=response_model,
                             tags=[bp_name])(
                        create_endpoint_function(other_params)
                    )
                elif other_params["method"] == 'put':
                    app.put(endpoint_obj.route, response_model=response_model,
                            tags=[bp_name])(
                        create_endpoint_function(other_params)
                    )
                elif other_params["method"] == 'delete':
                    app.delete(endpoint_obj.route, response_model=response_model,
                               tags=[bp_name])(
                        create_endpoint_function(other_params)
                    )


def generic_route_handler(
    # *args,
    **kwargs,
) -> Response:
    """
    Handles generic route requests and delegates to the appropriate
    CRUD operation based on the request parameters.

    Args:
        kwargs (dict): Additional keyword arguments.

    Returns:
        Response: The response from the CRUD operation.
    """
    other_params = kwargs["other_params"]
    request = kwargs["request"]
    blueprint = kwargs["blueprint"]

    if DEBUG:
        log_debug(
            "generic_route_handler |" +
            f" | kwargs: {kwargs}"
        )
        pprint(request.to_dict())

    if DEBUG:
        log_debug(
            "generic_route_handler |" +
            f" other_params: {other_params}" +
            f" request: {request}"
        )

    # Set environment variables from the database configurations.
    app_context = app_context_and_set_env(request=request, blueprint=blueprint)
    if app_context.has_error():
        return return_resultset_jsonified_or_exception(
            app_context.get_error_resultset()
        )

    ep_helper = GenericEndpointHelper(
        app_context=app_context,
        json_file=other_params['params']["json_file"],
        url_prefix=other_params["name"]
    )
    if ep_helper.dbo.table_type == "child_listing" \
       and ep_helper.dbo.sub_type == "array":
        return ep_helper.generic_array_crud()
    return ep_helper.generic_crud_main()


def create_endpoint_function(other_params: dict) -> callable:
    """
    Creates a function that can be used as a FastAPI endpoint.

    Args:
        other_params (dict): The other parameters to pass to the endpoint function.

    Returns:
        callable: The endpoint function.
    """
    router = BlueprintOne()

    async def generic_get(
        request: FaRequest,
        current_user: str = Depends(get_current_user),
        _id: str = Query(None, min_length=3, max_length=50, alias="id"),
        user_id: Union[str, None] = Query(None),
        limit: Union[int, None] = Query(None, ge=1, le=100000),
        page: Union[int, None] = Query(None, ge=1, le=100000),
        like_param: str = Query(None),
        comb_param: str = Query(None),
    ):
        """
        Handles generic GET requests, to get one row by id, a Like Search,
        get one row by additional_query_params, or fetch the
        paginated-limited-filtered list.

        Args:
            id (str): The ID of the record to retrieve.
            limit (int): The maximum number of records to retrieve.
            page (int): The page number to retrieve.
            like_param (str): The parameter to search for records.
            comb_param (str): The parameter to combine with the like_param.

        Returns:
            Response: The response from the GET operation.
        """
        gs_request = build_request(
            method="get",
            query_params={
                "id": _id,
                "limit": limit,
                "page": page,
                "like_param": like_param,
                "comb_param": comb_param,
                "user_id": user_id,
            },
            headers={
                "current_user": current_user,
            }
        )
        router.set_current_request(request, gs_request)
        return generic_route_handler(other_params=other_params,
            request=gs_request, blueprint=router)

    async def generic_post(
        request: FaRequest,
        current_user: str = Depends(get_current_user),
        json_body: dict = Body(...),
    ):
        """
        Handles generic POST requests, to create one item.

        Args:
            json_body (dict) = data for the item to be created

        Returns:
            Response: The response from the GET operation.
        """
        gs_request = build_request(
            method="post",
            json_body=json_body,
            headers={
                "current_user": current_user,
            },
        )
        router.set_current_request(request, gs_request)
        return generic_route_handler(other_params=other_params,
            request=gs_request, blueprint=router)

    async def generic_put(
        request: FaRequest,
        current_user: str = Depends(get_current_user),
        json_body: dict = Body(...),
        update_item: str = Query(None),
    ):
        """
        Handles generic PUT requests, to update one item.

        Args:
            json_body (dict) = data to be updated
            update_item (str): "1" apply update_one() instead
                    of replace_one(). Defaults to "0"

        Returns:
            Response: The response from the GET operation.
        """
        gs_request = build_request(
            method="put",
            query_params={
                "update_item": update_item,
            },
            json_body=json_body,
            headers={
                "current_user": current_user,
            },
        )
        router.set_current_request(request, gs_request)
        return generic_route_handler(other_params=other_params,
            request=gs_request, blueprint=router)

    async def generic_delete(
        request: FaRequest,
        current_user: str = Depends(get_current_user),
        _id: str = Query(None, min_length=3, max_length=50, alias="id"),
        json_body: Optional[Union[dict, None]] = Body(...),
    ):
        """
        Handles generic DELETE requests, to delete one row by id.

        Args:
            id (str): The ID of the record to retrieve.

        Returns:
            Response: The response from the GET operation.
        """
        gs_request = build_request(
            method="delete",
            query_params={
                "id": _id,
            },
            json_body=json_body,
            headers={
                "current_user": current_user,
            },
        )
        router.set_current_request(request, gs_request)
        return generic_route_handler(other_params=other_params,
            request=gs_request, blueprint=router)

    if other_params["method"] == 'post':
        return generic_post
    if other_params["method"] == 'put':
        return generic_put
    if other_params["method"] == 'delete':
        return generic_delete
    # Defaults to other_params["method"] == 'get'
    return generic_get



"""
Menu options access
"""
from typing import Optional, Any

from genericsuite.util.framework_abs_layer import Response

from genericsuite.util.jwt import (
    AuthorizedRequest
)
from genericsuite.util.security import (
    get_authorized_menu_options,
    get_option_access,
)
from genericsuite.util.utilities import (
    get_request_body,
    return_resultset_jsonified_or_exception,
    get_default_resultset,
)
from genericsuite.config.config_from_db import app_context_and_set_env


def menu_options_get(
    request: AuthorizedRequest,
    blueprint: Any,
    other_params: Optional[dict] = None
) -> Response:
    """ Get authorized menu options """
    if other_params is None:
        other_params = {}
    # Set environment variables from the database configurations.
    app_context = app_context_and_set_env(request=request, blueprint=blueprint)
    if app_context.has_error():
        return return_resultset_jsonified_or_exception(
            app_context.get_error_resultset()
        )
    # Get authorized menu options
    menu_response, _, _ = get_authorized_menu_options(
        app_context=app_context, url_prefix=__name__
    )
    return return_resultset_jsonified_or_exception(
        menu_response
    )


def menu_options_element(
    request: AuthorizedRequest,
    blueprint: Any,
    other_params: Optional[dict] = None
) -> Response:
    """ Get menu element configuration """
    if other_params is None:
        other_params = {}
    # Set environment variables from the database configurations.
    # app_context = app_context_and_set_env(request)
    app_context = app_context_and_set_env(request=request, blueprint=blueprint)
    if app_context.has_error():
        return return_resultset_jsonified_or_exception(
            app_context.get_error_resultset()
        )
    # Read parameters
    request = app_context.get_request()
    params = get_request_body(request)
    element = params.get('element')
    oa_response = get_default_resultset()
    if element is None:
        oa_response['error'] = True
        oa_response['error_message'] = "'element' parameter must" + \
            " be specified [MOEEP-010]"
        return return_resultset_jsonified_or_exception(
            oa_response, 403
        )
    # Get menu element configuration
    oa_response = get_option_access(app_context, element)
    if oa_response['error']:
        # Access denied
        return return_resultset_jsonified_or_exception(
            oa_response, 401
        )
    # Access granted
    oa_response['response'] = True
    return return_resultset_jsonified_or_exception(oa_response)



"""
App main module (create_app) for Chalice
"""
from typing import Any
import importlib
import logging

from genericsuite.util.app_logger import log_info
# from genericsuite.util.app_logger import log_debug

from genericsuite.config.config import Config

from genericsuite.chalicelib.endpoints import users
from genericsuite.chalicelib.endpoints import menu_options
from genericsuite.chalicelib.endpoints import storage_retrieval
from genericsuite.chalicelib.util.generic_endpoint_builder import (
    generate_blueprints_from_json
)

from genericsuite.config.config_from_db import set_init_custom_data

# from chalice import Chalice
framework_class = importlib.import_module("chalice")

DEBUG = False


def create_app(app_name: str, settings=None) -> Any:
    """ Create the Chalice App """

    if settings is None:
        settings = Config()

    chalice_app = framework_class.Chalice(app_name=app_name)
    chalice_app.experimental_feature_flags.update(['BLUEPRINTS'])

    chalice_app.debug = settings.DEBUG

    # Custom data, to be used to store a dict for
    # GenericDbHelper specific functions
    chalice_app.custom_data = set_init_custom_data()

    # App wide log level
    if not settings.DEBUG:
        chalice_app.log.setLevel(logging.INFO)

    # CORS configuration
    chalice_app.api.cors = set_cors_config(
        cors_config_class=framework_class.CORSConfig,
        settings=settings)

    # Set Content-type: multipart/form-data as Binary
    # to properly handle image uploads
    chalice_app.api.binary_types.append("multipart/form-data")
    # log_debug(f'api.binary_types: {chalice_app.api.binary_types}')

    # Register general endpoints
    chalice_app.register_blueprint(menu_options.bp, url_prefix='/menu_options')
    # log_info("Registered menu_options blueprint")
    chalice_app.register_blueprint(users.bp, url_prefix='/users')
    # log_info("Registered users blueprint")
    chalice_app.register_blueprint(storage_retrieval.bp, url_prefix='/asset')
    # log_info("Registered storage_retrieval blueprint")

    # Register generic endpoints (from the "endpoints.json" file)
    generate_blueprints_from_json(chalice_app, 'endpoints')

    # Anounce app boot
    log_info(f"{settings.APP_NAME.capitalize()} v" +
             settings.APP_VERSION +
             " | Accepting connections from: " +
             settings.CORS_ORIGIN)
    if DEBUG:
        log_info(settings.debug_vars())

    return chalice_app


def set_cors_config(cors_config_class, settings):
    """
    Sets the CORS configuration for the API.
    Returns:
        CORSConfig: The CORS configuration.
    """
    cors_config = cors_config_class(
        allow_origin=settings.CORS_ORIGIN,
        allow_headers=[
            settings.HEADER_TOKEN_ENTRY_NAME,
            'Access-Control-Allow-Origin',
            'Content-Type',
            'Access-Control-Allow-Headers',
        ],
        max_age=600,
        expose_headers=[
            settings.HEADER_TOKEN_ENTRY_NAME,
            'Access-Control-Allow-Origin',
            'Content-Type',
            'Content-Disposition',
        ],
        allow_credentials=True
    )
    return cors_config



"""
System users operations (CRUD, login, database test, super-admin creation)
"""
from fastapi import Request as FaRequest, Depends
# from fastapi.security import OAuth2PasswordRequestForm
from fastapi.security import HTTPBasic, HTTPBasicCredentials

from genericsuite.util.framework_abs_layer import Response
from genericsuite.fastapilib.util.blueprint_one import BlueprintOne
from genericsuite.fastapilib.util.dependencies import (
    get_current_user,
    get_default_fa_request,
)
from genericsuite.models.users.users import (
    test_connection_handler as test_connection_handler_model,
    login_user as login_user_model,
    super_admin_create as super_admin_create_model,
    get_current_user_data,
)

# router = APIRouter()
router = BlueprintOne()

# Set up Basic Authentication
security = HTTPBasic()

HEADER_CREDS_ENTRY_NAME = 'Authorization'
DEBUG = False


@router.get('/test', tags='test')
async def test_connection_handler(
    request: FaRequest,
    current_user: str = Depends(get_current_user),
) -> Response:
    """Connection handler test"""
    gs_request, other_params = get_default_fa_request(current_user)
    router.set_current_request(request, gs_request)
    return test_connection_handler_model(gs_request, other_params)


@router.post('/login', tags='login')
async def login_user(
    request: FaRequest,
    # form_data: OAuth2PasswordRequestForm = Depends()
    credentials: HTTPBasicCredentials = Depends(security)
) -> Response:
    """User login"""
    gs_request, other_params = get_default_fa_request()
    router.set_current_request(request, gs_request)
    other_params['username'] = credentials.username
    other_params['password'] = credentials.password
    return login_user_model(
        request=gs_request, blueprint=router,
        other_params=other_params)


@router.post('/supad-create', tags='super-admin')
async def super_admin_create(
    request: FaRequest,
    # form_data: OAuth2PasswordRequestForm = Depends()
    credentials: HTTPBasicCredentials = Depends(security)
) -> Response:
    """Super admin user emergency creation"""
    gs_request, other_params = get_default_fa_request()
    router.set_current_request(request, gs_request)
    other_params['username'] = credentials.username
    other_params['password'] = credentials.password
    return super_admin_create_model(
        request=gs_request, blueprint=router,
        other_params=other_params)


@router.get('/current_user_d')
async def current_user_d(
    request: FaRequest,
    current_user: str = Depends(get_current_user),
) -> Response:
    """
    Current user data read
    """
    gs_request, other_params = get_default_fa_request(current_user)
    router.set_current_request(request, gs_request)
    return get_current_user_data(gs_request, router, other_params)



"""
System users operations (CRUD, login, database test, super-admin creation)
"""
from typing import Optional, Any
import json

from genericsuite.util.framework_abs_layer import Request, Response

from genericsuite.util.db_abstractor import (
    set_db_request,
    test_connection,
)
from genericsuite.util.generic_db_helpers import (
    GenericDbHelper,
)
from genericsuite.util.app_logger import log_debug
from genericsuite.util.jwt import (
    token_encode,
    get_basic_auth,
    AuthorizedRequest,
)
from genericsuite.util.passwords import Passwords
from genericsuite.util.utilities import (
    return_resultset_jsonified_or_exception,
    get_default_resultset,
    get_id_as_string,
)
from genericsuite.config.config import Config
from genericsuite.util.generic_endpoint_helpers import GenericEndpointHelper
from genericsuite.config.config_from_db import app_context_and_set_env
from genericsuite.constants.const_tables import get_constant


HEADER_CREDS_ENTRY_NAME = 'Authorization'
DEBUG = False


def users_crud(
    request: AuthorizedRequest,
    blueprint: Any,
    other_params: Optional[dict] = None
) -> Response:
    """ User's CRUD operations (create, read, update, delete) """
    if not other_params:
        other_params = {}
    # Set environment variables from the database configurations.
    app_context = app_context_and_set_env(request=request, blueprint=blueprint)
    if app_context.has_error():
        return return_resultset_jsonified_or_exception(
            app_context.get_error_resultset()
        )
    # User's CRUD operations
    ep_helper = GenericEndpointHelper(
        app_context=app_context,
        json_file="users",
        url_prefix=__name__,
    )
    return ep_helper.generic_crud_main()


def test_connection_handler(
    request: Request,
    other_params: Optional[dict] = None
) -> Response:
    """Connection handler test"""
    if not other_params:
        other_params = {}
    result = get_default_resultset()
    set_db_request(request)
    result['resultset']['collections'] = json.loads(test_connection())
    if DEBUG:
        log_debug(f'Test DB connection | request: {request}')
    return return_resultset_jsonified_or_exception(result)


def login_user(
    request: Request,
    blueprint: Any,
    other_params: Optional[dict] = None
) -> Response:
    """User login"""
    if not other_params:
        other_params = {}
    psw_class = Passwords()
    dbo = GenericDbHelper(json_file="users", request=request,
                          blueprint=blueprint)
    if DEBUG:
        log_debug(f'login_user | request: {request}')
        # log_debug('login_user | bp.current_request.to_dict(): ' +
        #           f'{bp.current_request.to_dict()}')
    result = get_default_resultset()
    if other_params.get('username') and other_params.get('password'):
        username = other_params.get('username')
        password = other_params.get('password')
    else:
        basic_auth_data = get_basic_auth(
            request.headers.get(HEADER_CREDS_ENTRY_NAME, '')
        )
        if DEBUG:
            log_debug(f'login_user | basic_auth_data: {basic_auth_data}')
        if basic_auth_data['error']:
            # 'Could not verify [L1]'
            result['error_message'] = basic_auth_data['error_message']
            return return_resultset_jsonified_or_exception(
                result=result,
                http_error=basic_auth_data['status_code']
            )
        username = basic_auth_data['resultset']['user']
        password = basic_auth_data['resultset']['password']

    user = dbo.fetch_row_by_entryname_raw('email', username)
    if user['error']:
        return return_resultset_jsonified_or_exception(user)
    _ = DEBUG and \
        log_debug(f'login_user | user: {user}')
    if user['resultset']:
        if 'passcode' in user['resultset']:
            if psw_class.verify_password(user['resultset']['passcode'],
                                         password):

                if user['resultset'].get('status', '1') == '1':
                    token = token_encode(user['resultset'])
                    result['resultset'] = {
                        'token': token,
                        '_id': get_id_as_string(user['resultset']),
                        'firstname': user['resultset']['firstname'],
                        'lastname': user['resultset']['lastname'],
                        'email': user['resultset']['email'],
                        'username': user['resultset']['email'],
                    }
                else:
                    result['error_message'] = get_constant("ERROR_MESSAGES",
                                                           "ACCOUNT_INACTIVE")
            else:
                result['error_message'] = 'Could not verify [L3]'
        else:
            result['error_message'] = 'Inconsistency [L4]'
    else:
        result['error_message'] = 'Could not verify [L2]'
    _ = DEBUG and \
        log_debug(f'login_user | FINAL result: {result}')
    return return_resultset_jsonified_or_exception(result, 401)


def super_admin_create(
    request: Request,
    blueprint: Any,
    other_params: Optional[dict] = None
) -> Response:
    """Super admin user emergency creation"""
    # Set environment variables from the database configurations.
    app_context = app_context_and_set_env(request=request, blueprint=blueprint)
    if app_context.has_error():
        return return_resultset_jsonified_or_exception(
            app_context.get_error_resultset()
        )
    request = app_context.get_request()
    settings = Config(app_context)
    if not other_params:
        other_params = {}
    psw_class = Passwords()
    dbo = GenericDbHelper(json_file="users", request=request, blueprint=blueprint)
    result = get_default_resultset()

    if other_params.get('username') and other_params.get('password'):
        username = other_params.get('username')
        password = other_params.get('password')
    else:
        basic_auth_data = get_basic_auth(
            request.headers.get(HEADER_CREDS_ENTRY_NAME, '')
        )
        if DEBUG:
            log_debug(f'supad-create | basic_auth_data: {basic_auth_data}')
        if basic_auth_data['error']:
            result['error_message'] = basic_auth_data['error_message']
            return return_resultset_jsonified_or_exception(
                result=result,
                http_error=basic_auth_data['status_code']
            )

        username = basic_auth_data['resultset']['user']
        password = basic_auth_data['resultset']['password']

    if not username or not password:
        result['error_message'] = 'Could not verify [SAC1]'
    elif username != settings.APP_SUPERADMIN_EMAIL:
        result['error_message'] = 'Could not verify [SAC2]'
    elif not psw_class.verify_password(
         psw_class.encrypt_password(settings.APP_SECRET_KEY), password
         ):
        result['error_message'] = 'Could not verify [SAC3]'

    if result['error_message']:
        return return_resultset_jsonified_or_exception(result)

    user = dbo.fetch_row_by_entryname_raw('email', username)
    if user['error']:
        return return_resultset_jsonified_or_exception(user)

    if user['resultset']:
        result['error_message'] = 'User already exists [SAC4]'
    else:
        request_body = {
            "firstname": "Admin",
            "lastname": "Super",
            "superuser": "1",
            "status": "1",
            "plan": "premium",
            "language": "en",
            'email':  username,
            'passcode': password,
            "creation_date": 1635033994,
            "update_date": 1635033994,
            "birthday": -131760000,
            "gender": 'm',
        }
        for field in dbo.get_mandatory_fields(record=request_body,
                                              is_create=True):
            if field not in request_body:
                request_body[field] = ''
        result = dbo.create_row(request_body)

    return return_resultset_jsonified_or_exception(result)


def get_current_user_data(
    request: AuthorizedRequest,
    blueprint: Any,
    other_params: Optional[dict] = None
) -> Response:
    """
    Current user data read
    """
    if not other_params:
        other_params = {}
    app_context = app_context_and_set_env(request=request, blueprint=blueprint)
    if app_context.has_error():
        return return_resultset_jsonified_or_exception(
            app_context.get_error_resultset()
        )
    result = get_default_resultset()
    result['resultset'] = app_context.get_user_data()
    return return_resultset_jsonified_or_exception(result)



"""
The parse_multipart module handles the multipart form-data parsing,
and get content like upoaded files.
"""
import os

from requests_toolbelt import MultipartDecoder

from genericsuite.util.app_logger import log_debug
from genericsuite.util.utilities import return_resultset_jsonified_or_exception
from genericsuite.util.app_context import AppContext
from genericsuite.util.file_utilities import temp_filename

DEBUG = False


def parse_multipart(raw_body, headers):
    """
    The parse_multipart function is implemented to handle the multipart
    form data parsing, and get content like upoaded files.
    (solves "multipart/form-data" and Chalice nightmare!)
    """
    content_type = headers['content-type']
    if content_type == 'multipart/form-data':
        content_type += "; boundary=----WebKitFormBoundary7MA4YWxkTrZu0gW"
    decoder = MultipartDecoder(raw_body, content_type)

    # It returns the list of parts, but only content and encoding.
    # For files, he "name" (and "file_name") weont be passed.

    def get_part(part):
        disposition = part.headers.get(b'Content-Disposition', "")
        params = {}
        for disp_part in str(disposition).split(';'):
            kv = disp_part.split('=', 2)
            params[str(kv[0]).strip()] = (
                str(kv[1]).strip('\"\'\t \r\n')
                if len(kv) > 1 else str(kv[0]).strip()
            )
        part_type = (
            part.headers[b'Content-Type']
            if b'Content-Type' in part.headers else None
        )
        return {
            "content": part.content,
            "type": part_type,
            "params": params,
            "headers": part.headers,
            "encoding": part.encoding,
        }

    # parsed_parts = {"parts": [p.content for p in decoder.parts]}
    parsed_parts = {
        "parts": [get_part(p) for p in decoder.parts]
    }
    if DEBUG:
        log_debug("")
        log_debug(f">>--> PARSE_MULTIPART | parsed_parts: {parsed_parts}")
        log_debug("")
    return parsed_parts


def file_upload_handler(app_context: AppContext, p: dict):
    """
    Handle the file upload process.

    This function takes a dictionary with 'extension' and 'handler_function',
    saves the uploaded file to a temporary directory, processes it using the
    specified handler function, and then cleans up by removing the temporary
    file.

    :param p: A dictionary containing 'extension' and 'handler_function'.
    :return: The result of the handler function or an error message.
    """
    request = app_context.get_request()
    if "other_params" not in p:
        p["other_params"] = {}

    if "uploaded_file_path" in p and p.get("uploaded_file_path"):
        # File was already uploaded (e.g. FastAPI)
        if p['uploaded_file_path']:
            file_path = p['uploaded_file_path']
        else:
            return return_resultset_jsonified_or_exception({
                'error': True,
                'error_message': 'No file provided [1]'
            })
    else:
        # Ensure the request content type is multipart/form-data (e.g. Chalice)
        if request.headers['Content-Type'].startswith('multipart/form-data'):
            parsed_body = parse_multipart(request.raw_body, request.headers)

            if parsed_body['parts'][0]["content"]:
                file_path = download_file(
                    parsed_body['parts'][0]["content"],
                    p['extension'],
                )
            else:
                return return_resultset_jsonified_or_exception({
                    'error': True,
                    'error_message': 'No file provided [2]'
                })
        else:
            return return_resultset_jsonified_or_exception({
                'error': True,
                'error_message': 'Unsupported media type, expected' +
                ' multipart/form-data'
            })

    # Pass additional parameters from the 'other_params' dict to
    # the handler function
    param_name = p.get("file_path_param_name", "file_path")
    if "unique_param_name" in p:
        p["other_params"][p["unique_param_name"]][param_name] = \
            file_path
    else:
        p["other_params"][param_name] = file_path

    # Call the handler function
    log_debug('Call the handler function:\n' +
                f'{p["handler_function"]}({p["other_params"]})')

    result = p["handler_function"](
        **p["other_params"]
    )
    if p.get("delete_file_after_processing", True):
        if DEBUG:
            log_debug(f">> Cleaning up: {file_path}...")
        os.remove(file_path)  # Clean up the temporary file
    else:
        log_debug(f">> File left in device: {file_path}")
    return return_resultset_jsonified_or_exception(result)


def download_file(file_data: bytes, extension: str) -> str:
    """
    Download the file data to a temporary directory.

    :param file_data: The file data to download.
    :return: The path to the downloaded file.
    """
    file_path = temp_filename(extension)
    with open(file_path, 'wb') as file:
        file.write(file_data)
    return file_path



"""
Blueprint wrapper to add authorization, other data and schema validation
to requests.
"""
from typing import Optional, Any

from fastapi import FastAPI
from fastapi import APIRouter, Request

# from genericsuite.util.app_logger import log_debug


class BlueprintOne(APIRouter):
    """
    Class to register a new route with optional schema validation and authorization.
    """
    name: str = "NoName"
    current_app: Optional[Any] = None
    current_request: Optional[Any] = None
    current_fa_request: Optional[Any] = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if len(args) > 0:
            self.name = args[0]

    def get_current_app(self) -> Any:
        """
        Get the current App object. It must be called inside a router function.
        """
        # self.current_request = Request(scope={"type": "http"})
        # self.current_app = self.app
        # log_debug(f">> BlueprintOne(APIRouter) | Current App: {self.current_app}")
        return self.current_app

    def set_current_app(self, app: FastAPI) -> Any:
        """
        Get the current App object. It must be called inside a router function.
        """
        self.current_app = app
        return self.current_app

    def get_current_fa_request(self) -> Any:
        """
        Get the current App object. It must be called inside a router function.
        """
        # self.current_request = Request(scope={"type": "http"})
        return self.current_fa_request

    def get_current_request(self) -> Any:
        """
        Get the current App object. It must be called inside a router function.
        """
        # self.current_request = Request(scope={"type": "http"})
        return self.current_request

    def set_current_request(self, fa_request: Request, gs_request: Any) -> Any:
        """
        Get the current App object. It must be called inside a router function.
        """
        self.current_request = gs_request
        self.current_fa_request = fa_request
        self.set_current_app(fa_request.app)
        return self.current_request



"""
Context manager to preserve data between GPT functions
"""
from typing import Any, Union, Optional, Callable
import os
import json

from genericsuite.config.config_secrets import get_secrets_cache_filename
from genericsuite.constants.const_tables import get_constant
from genericsuite.util.current_user_data import (
    get_curr_user_data,
    get_curr_user_id,
    NON_AUTH_REQUEST_USER_ID,
)
from genericsuite.util.jwt import AuthorizedRequest
from genericsuite.util.utilities import get_default_resultset, get_id_as_string
from genericsuite.util.app_logger import log_debug


DEBUG = False
TEMP_DIR = os.environ.get('TEMP_DIR', '/tmp')

PARAMS_FILE_ENABLED = os.environ.get('PARAMS_FILE_ENABLED', '1')
PARAMS_FILE_USER_FILENAME_TEMPLATE = os.environ.get(
    'PARAMS_FILE_USER_FILENAME_TEMPLATE', 'params_[user_id].json')
PARAMS_FILE_GENERAL_FILENAME = os.environ.get(
    'PARAMS_FILE_GENERAL_FILENAME', 'params_GENERAL.json')


class ParamsFile():
    """
    Class to manage the parameters file (/tmp/param_*.json)
    """
    def __init__(self, user_id: str):
        self.user_id = user_id

    def get_params_file_path(self, filename: str):
        """
        Get the path of the file.
        """
        return os.path.join(TEMP_DIR, filename)

    def get_params_filename(self, user_id: Optional[str] = None
                            ) -> Union[str, None]:
        """
        Get the filename where the parameters are stored.
        """
        if not user_id:
            user_id = self.user_id
        if user_id == NON_AUTH_REQUEST_USER_ID:
            # For the un-authenticated endpoint calls file
            filename = None
        else:
            filename = self.get_params_file_path(
                PARAMS_FILE_USER_FILENAME_TEMPLATE.replace(
                    '[user_id]', user_id))
        _ = DEBUG and \
            log_debug(
                'GET_FILENAME-1) get_params_filename |' +
                f' user_id: {user_id} | filename: {filename}')
        return filename

    def load_params_file(self, filename: str) -> dict:
        """
        Load the parameters from a file.
        """
        result = get_default_resultset()
        result['found'] = True
        if PARAMS_FILE_ENABLED != '1':
            result['error_message'] = "Params. file flag disabled"
            result['found'] = False
        elif not filename:
            result['error_message'] = "Filename is null"
            result['found'] = False
        elif not os.path.exists(filename):
            result['error_message'] = f"Filename does not exist: {filename}"
            result['found'] = False
        if result['found']:
            with open(filename, 'r', encoding='utf-8') as fhdlr:
                result['resultset'] = json.load(fhdlr)
        _ = DEBUG and \
            log_debug(f'LOAD_PF-1) load_params_file | File: {filename}' +
                      f' | {result["resultset"]}')
        return result

    def save_params_file(self, filename: str, data_to_save: dict) -> dict:
        """
        Store the parameters in a file.
        """
        result = get_default_resultset()
        result['resultset'] = data_to_save
        if PARAMS_FILE_ENABLED != '1':
            return result
        if '_id' in data_to_save:
            data_to_save['_id'] = get_id_as_string(data_to_save)
        with open(filename, 'w', encoding='utf-8') as f:
            json.dump(data_to_save, f)
        _ = DEBUG and log_debug('PF-2) save_params_file |' +
                                f' filename: {filename} |' +
                                f' content: {data_to_save}')
        return result


class AppContext:
    """
    Context manager class to preserve data between GPT functions
    """
    def __init__(
        self,
        request: Optional[Union[AuthorizedRequest, None]] = None,
        blueprint: Optional[Any] = None
    ):
        # AuthorizedRequest object received from the endpoit handler
        self.request = request
        # Blueprint object received from the endpoit handler
        self.blueprint = blueprint
        # Curret user's data
        self.user_data = None
        # Any error happened during the user's data or other retrievals
        self.error_message = None
        # "self.other" was created to make globally available some values that
        # should not be calculated each time a GPT function is called.
        self.other_data = {}
        # Environment variables loaded from the database
        self.env_data = {}

    def set_context_from_blueprint(self, blueprint: Any, request: Any = None):
        """ Set the request object """
        self.blueprint = blueprint
        if request:
            self.set_context(request)
        else:
            self.set_context(blueprint.get_current_request())

    def set_context(self, request: Any):
        """ Set the request object """
        self.request = request
        self.user_data = self.get_user_data()

    def get_user_id(self):
        """ Get current user ID """
        return get_curr_user_id(self.request)

    def get_request(self):
        """ Get the request object """
        return self.request

    def get_blueprint(self):
        """ Get the blueprint object """
        return self.blueprint

    def get_user_data_raw(self) -> None:
        """
        Get current user data
        """
        # user_response = get_curr_user_data(self.request)
        user_response = get_curr_user_data(
            request=self.request,
            blueprint=self.blueprint)
        if user_response['error']:
            self.set_error(user_response['error_message'])
            return
        self.user_data = user_response['resultset']
        if not self.user_data:
            if 'PYTEST_CURRENT_TEST' in os.environ:
                self.user_data = {
                    "_id": self.get_user_id(),
                }
            elif self.get_user_id() == NON_AUTH_REQUEST_USER_ID:
                # It's a normal Request without Authorization headers
                self.user_data = {}
            else:
                # Inconsistency Error
                self.set_error(get_constant(
                    "ERROR_MESSAGES",
                    "NO_USER_DATA", "System Error [AC-GUD-E010]" +
                    f" | id: {self.get_user_id()}"))
                self.user_data = {}
        elif self.user_data.get('status', '1') == '0':
            self.set_error(get_constant(
                "ERROR_MESSAGES",
                "ACCOUNT_INACTIVE", "User account inactive [AC-GUD-E020]"))

    def get_user_data(self, check_params_file: Optional[bool] = True):
        """
        Get current user data verifying the params json cache file existence
        """
        if not self.user_data:
            if check_params_file and PARAMS_FILE_ENABLED == '1':
                pfc = ParamsFile(self.get_user_id())
                filename = pfc.get_params_filename()
                if not filename:
                    self.get_user_data_raw()
                else:
                    params_file_result = pfc.load_params_file(filename)
                    if params_file_result['found']:
                        # self.user_data = \
                        #   params_file_result['resultset']['user_data']
                        self.user_data = params_file_result['resultset']
                    else:
                        self.get_user_data_raw()
                        if '_id' in self.user_data:
                            pfc.save_params_file(filename, self.user_data)
            else:
                self.get_user_data_raw()
        _ = DEBUG and log_debug(
            "AppContext | GET_USER_DATA" +
            f"\n | self.user_data: {self.user_data}" +
            f"\n | self.error_message: {self.error_message}")
        return self.user_data

    def get_other_data(
        self,
        element_name: str,
        generator_func: Callable = None,
        generator_params: dict = None,
    ):
        """
        Get "other" data element value.
        If it doesn't exist try to call the generator_func
        with generator_params to assign the element's value.

        Args:
            element_name (str): name of the element to retrieve.
            generator_func (callable): function to generate the element value.
            The function should return a
            generator_params (dict): parameters for the generator fuction.

        Retuns:
            Any: the element's value or None if it's not present.
        """
        if element_name not in self.other_data:
            if generator_func is None:
                return None
            result = generator_func(**generator_params)
            if result.get("error", False):
                self.other_data[element_name] = result
            else:
                self.set_error(result.get(
                    "error_message", "ERROR calling generator_func [GPTC-010]"
                ))
        return self.other_data[element_name]

    def set_error(self, error_message: str) -> None:
        """
        Assign the internal error message
        """
        self.error_message = error_message

    def get_error(self) -> str:
        """
        Get the internal error message
        """
        return self.error_message

    def has_error(self) -> str:
        """
        Returns True if there's any error
        """
        return self.error_message is not None

    def get_error_resultset(self) -> dict:
        """
        Get the internal error message as a standard resultset
        """
        resultset = get_default_resultset()
        resultset["error"] = True
        resultset["error_message"] = self.error_message
        return resultset

    def set_other_data(
        self,
        element_name: str,
        value: Any,
    ) -> Any:
        """ Get "other" data element value. """
        self.other_data[element_name] = value
        return self.other_data[element_name]

    def get_env_var(self, var_name: str, def_value: Any = None) -> Any:
        """ Get environment variable value """
        return self.env_data.get(var_name, def_value)

    def set_env_var(
        self,
        var_name: str,
        value: Any,
    ) -> Any:
        """ Set environment variable value """
        self.env_data[var_name] = value
        return self.env_data[var_name]


class CommonAppContext():
    """
    Common context manager for the Langchain AI Chatbot API
    """
    def __init__(self):
        self.app_context = None

    def set(self, app_context: AppContext):
        """
        Set the common app context
        """
        self.app_context = app_context

    def get(self):
        """
        Get the common app context
        """
        return self.app_context


def get_app_context(app_context_or_blueprint: Any):
    """
    Get the application context object
    """
    if isinstance(app_context_or_blueprint, AppContext):
        app_context = app_context_or_blueprint
    else:
        app_context = AppContext()
        app_context.set_context_from_blueprint(app_context_or_blueprint)
    return app_context


def delete_params_file(
    app_context_or_blueprint: Any,
    action_data: Optional[Union[dict, None]]
) -> None:
    """
    GenericDbHelper specific function to delete the parameters file (e.g. when
    general_config or user's users_config array is changed).

    Args:
        app_context (AppContext): the application context object
        action_data (dict, optional): the action data. Defaults to None.
            If it's not None, it must have the following keys (attributes):
            "action": "list", "read", "create", "update" or "delete"
            "resultset": resultset for data to be stored, delete or
                retrieved with the keys: resultset, error, error_message.
            "cnf_db": the table configuration. E.g. tablename is
                cnf_db['tablename']
    """
    app_context = get_app_context(app_context_or_blueprint)
    pfc = ParamsFile(app_context.get_user_id())
    action_data = action_data or {}
    tablename = action_data.get("cnf_db", {}).get("table_name")

    _ = DEBUG and log_debug("AppContext | DELETE_PARAMS_FILE" +
                            f"\n | action_data: {action_data}" +
                            f"\n | tablename: {tablename}")
    # Only delete the params file if the action is not read or list
    if action_data.get("action") in ["read", "list"]:
        return action_data['resultset']
    # Verify if any error
    if action_data['resultset']['error']:
        return action_data['resultset']
    if tablename:
        # Get the filename according to the table name
        if tablename == 'general_config':
            filenames = [
                pfc.get_params_file_path(PARAMS_FILE_GENERAL_FILENAME),
                get_secrets_cache_filename("secrets"),
                get_secrets_cache_filename("envs"),
            ]
        else:
            # Get the user ID if it's not the general table

            # Get the database item resultset
            db_item = action_data['resultset']['resultset']
            if isinstance(db_item, str):
                db_item = json.loads(db_item)
            _ = DEBUG and log_debug(
                'DB SPECIFIC FUNCTION:' +
                f" delete_params_file | db_item: {db_item}")
            # Inconsistency If _id not in resultset...
            if '_id' not in db_item:
                return action_data['resultset']

            # Get the user ID from the resultset
            user_id = get_id_as_string(db_item)
            _ = DEBUG and log_debug(
                'DB SPECIFIC FUNCTION:' +
                f" delete_params_file | user_id: {user_id}")

            filenames = [pfc.get_params_filename(user_id)]

        _ = DEBUG and log_debug("AppContext | DELETE_PARAMS_FILE" +
                                f"\n | filenames: {filenames}")

        # Delete params file if exists
        for filename in filenames:
            if os.path.exists(filename):
                os.remove(filename)
                _ = DEBUG and log_debug("AppContext | DELETE_PARAMS_FILE" +
                                        f"\n | File deleted: {filename}")
            else:
                _ = DEBUG and log_debug("AppContext | DELETE_PARAMS_FILE" +
                                        f"\n | File not found: {filename}")
    return action_data['resultset']



"""
Date/time utilities
"""
from typing import Union, Any
from datetime import timedelta
from datetime import timezone
from datetime import datetime

from genericsuite.util.app_logger import log_debug, log_error

DEBUG = False


def current_datetime_timestamp() -> float:
    """Getting the current UTC date and time as a timestamp"""
    dt = datetime.now(timezone.utc)
    utc_time = dt.replace(tzinfo=timezone.utc)
    utc_timestamp = utc_time.timestamp()
    return utc_timestamp


def get_datetime_utc(float_timestamp: float):
    """
    Returns the given timestamp as UTC
    (Coordinated Universal Time) timestamp
    """
    utc_dt = datetime.utcfromtimestamp(float_timestamp)
    return utc_dt

def ts_to_ymd(tstamp: float, only_date: bool = False) -> str:
    """
    Converts a timestamp to a string in the format YYYY-MM-DD
    """
    utc_dt = get_datetime_utc(tstamp)
    if only_date:
        return utc_dt.strftime('%Y-%m-%d')
    return utc_dt.strftime('%Y-%m-%d %H:%M:%S')


def get_date_zero_hour(input_ts: Union[float, str]) -> float:
    """
    Returns the given timestamp as a zero hour timestamp
    """
    if isinstance(input_ts, str):
        tstamp = float(input_ts)
    else:
        tstamp = input_ts
    # Get the UTC datetime object from the given timestamp
    # utc_dt = datetime.utcfromtimestamp(tstamp)
    # return utc_dt.replace(hour=0, minute=0, second=0,
    #                       microsecond=0).timestamp()
    # return tstamp - (tstamp % 3600)
    ts_ymd = ts_to_ymd(tstamp, only_date=True) + " 00:00:00"
    return datetime.strptime(ts_ymd, '%Y-%m-%d %H:%M:%S').replace(
        tzinfo=timezone.utc).timestamp()


def get_date_eod(input_ts: Union[float, str]) -> float:
    """
    Returns the given timestamp as a zero hour timestamp
    """
    if isinstance(input_ts, str):
        tstamp = float(input_ts)
    else:
        tstamp = input_ts
    ts_ymd = ts_to_ymd(tstamp, only_date=True) + " 23:59:59"
    return datetime.strptime(ts_ymd, '%Y-%m-%d %H:%M:%S').replace(
        tzinfo=timezone.utc).timestamp()


def get_date_range_filter(v: str, other_entries: dict = None
                          ) -> dict:
    """
    Generates a date range filter for MongoDB queries.

    Args:
        v (str): A string containing either a single timestamp or two
                    timestamps separated by a comma, representing the start
                    and end dates for the filter.

    Returns:
        dict: A dictionary with '$lte' and '$gte' keys for the MongoDB
                query, representing the range of dates to filter by.
    """
    dates = [None, None]
    if "," in v:
        # Date range separated by comma
        date_filter = {}
        dates = v.split(",")
        # The first date is the start date
        if dates[0].strip() != '':
            date_filter['$gte'] = get_date_zero_hour(dates[0])
        # The second date is the end date
        if len(dates) > 1 and dates[1].strip() != '':
            end_date = get_date_zero_hour(dates[1])
        else:
            end_date = datetime.utcnow().timestamp()
        # date_filter['$lte'] = (get_datetime_utc(end_date) +
        #                        timedelta(hours=24)).timestamp()
        date_filter['$lte'] = get_date_eod(end_date)
    else:
        date_filter = {
            # Same day
            # '$lte': (get_datetime_utc(get_date_zero_hour(v)) +
            #          timedelta(hours=24)).replace(
            #          tzinfo=timezone.utc).timestamp(),
            '$lte': get_date_eod(v),
            '$gte': get_date_zero_hour(v),
        }
    if not other_entries:
        other_entries = {}
    date_filter.update(other_entries)
    # if DEBUG:
    log_debug(f"GET_DATE_RANGE_FILTER | v: {v}" +
        f"\n | dates: {dates}" +
        f"\n | dates[0]: {dates[0]}" +
        f"\n | dates[1]: {dates[1]}" +
        f"\n | date_filter: {date_filter}" +
        f"\n | date_filter['$gte']: {ts_to_ymd(date_filter['$gte'])}" +
        f"\n | date_filter['$lte']: {ts_to_ymd(date_filter['$lte'])}" +
        "\n")
    return date_filter


def interpret_any_date(any_date: Any) -> float:
    """
    Convert date to timestamp from .

    Args:
        any_date (Any): the date as timestamp, or string with
            format "Month day, year" or "YYYY-MM-DD"

    Returns:
        float: date timestamp or -1 if ValueError error.
    """
    if isinstance(any_date, float):
        date_timestamp = any_date
    else:
        if "-" in any_date:
            date_fmt = '%Y-%m-%d'
        else:
            date_fmt = '%B %d, %Y'
        try:
            date_timestamp = \
                datetime.strptime(any_date, date_fmt).replace(
                    hour=0, minute=0, second=0, microsecond=0
                ).timestamp()
        except ValueError as err:
            log_error(f"Failed to convert {any_date} to" +
                      f" timestamp: {str(err)}")
            date_timestamp = -1
    return date_timestamp



"""
FastAPI dependencies library
"""
from typing import Optional, Union

from fastapi import HTTPException, Depends, Request as FaRequest
from fastapi.security import OAuth2PasswordBearer

from genericsuite.fastapilib.framework_abstraction import (
    Request,
)
from genericsuite.util.jwt import (
    get_general_authorized_request,
    AuthorizedRequest,
    AuthTokenPayload,
)


oauth2_scheme = OAuth2PasswordBearer(tokenUrl="token")


def build_request(
    method: Optional[str] = None,
    query_params: Optional[dict] = None,
    json_body: Optional[dict] = None,
    headers: Optional[dict] = None,
    preserve_nones: bool = False,
) -> Union[Request, AuthorizedRequest]:
    """
    Builds the request query parameters from query_received and cleans it,
    leaving only the ones that are not None

    Args:
        method (str): HTTP method.
        query_params (dict): URL query parameters.
        json_body (dict): Request JSON body.
        headers (dict): Request headers.

    Returns:
        dict: The cleaned request query parameters.
    """
    query_params_reduced = {}
    if query_params:
        # Reduce query_params leaving only the not None items
        query_params_reduced = {k: v for k, v in query_params.items()
                                if v is not None or preserve_nones}
    headers_reduced = headers if headers else {}
    if "token" in headers_reduced:
        headers_reduced["Authorization"] = f"Bearer {headers_reduced['token']}"
        del headers_reduced["token"]
    if "current_user" in headers_reduced:
        new_request = AuthorizedRequest(
            method=method if method else "get",
            query_params=query_params_reduced,
            json_body=json_body if json_body else {},
            headers=headers_reduced,
            user=headers_reduced["current_user"])
    else:
        new_request = Request(
            method=method if method else "get",
            query_params=query_params_reduced,
            json_body=json_body if json_body else {},
            headers=headers_reduced)
    return new_request


async def get_current_user(token: str = Depends(oauth2_scheme)):
    """
    Verifies the JWT token and returns the current user.
    """
    request = build_request(headers={"token": token})
    auth_request = get_general_authorized_request(request)
    credentials_exception = HTTPException(
        status_code=401,
        detail="Could not validate credentials",
        headers={"WWW-Authenticate": "Bearer"},
    )
    # if auth_request.user is None:
    if not isinstance(auth_request, AuthorizedRequest):
        raise credentials_exception
    return auth_request.user


def get_default_fa_request(
    current_user: Optional[AuthTokenPayload] = None,
    json_body: Optional[dict] = None,
    headers: Optional[dict] = None,
    query_params: Optional[dict] = None,
    other_params: Optional[dict] = None,
    preserve_nones: bool = False,
):
    """
    Builds the default FA (FastAPI) Authentication request object.
    """
    params = {
        "json_body": json_body if json_body else {},
        "headers": headers if headers else {},
        "query_params": query_params if query_params else {},
        "preserve_nones": preserve_nones,
    }
    if current_user:
        params["headers"]["current_user"] = current_user
    request = build_request(**params)
    other_params = other_params if other_params else {}
    return request, other_params



"""
Storage retrieval for FastAPI
"""
from typing import Union, Optional
from typing import Any
import os
import io

from fastapi import Request as FaRequest
from fastapi import BackgroundTasks
from fastapi.responses import StreamingResponse
from fastapi.responses import FileResponse

from genericsuite.util.framework_abs_layer import Request, Response
from genericsuite.fastapilib.util.blueprint_one import BlueprintOne
from genericsuite.fastapilib.util.dependencies import (
    get_default_fa_request,
)
from genericsuite.util.aws import storage_retieval
from genericsuite.util.app_logger import log_debug
from genericsuite.util.utilities import (
    return_resultset_jsonified_or_exception,
    send_file_text_text,
)
# from genericsuite.config.config import Config

DEBUG = True

DEFAULT_DOWNLOAD_METHOD = "fastapi"
# DEFAULT_DOWNLOAD_METHOD = "inline"
# DEFAULT_DOWNLOAD_METHOD = "streaming"
# DEFAULT_DOWNLOAD_METHOD = "attachment"

# router = APIRouter()
router = BlueprintOne()


@router.get('/')
async def storage_retrieval_no_item_id_endpoint(
    request: FaRequest,
    background_tasks: BackgroundTasks,
) -> Any:
    """ Get authorized menu options """
    gs_request, other_params = get_default_fa_request()
    router.set_current_request(request, gs_request)
    # Report the error ASR-E1010
    item_id = None
    return storage_retieval_fa(request=gs_request, blueprint=router,
        item_id=item_id, other_params=other_params,
        background_tasks=background_tasks)

# Why these endpoint definitions return type is "-> Any": and not:
# "-> Union[Response, FileResponse, StreamingResponse]":
#
# Because this error:
#
# fastapi.exceptions.FastAPIError: Invalid args for response field! Hint: check
# that typing.Union[genericsuite.util.framework_abs_layer.Response,
# starlette.responses.FileResponse, starlette.responses.StreamingResponse] is a
# valid Pydantic field type. If you are using a return type annotation that is
# not a valid Pydantic field (e.g. Union[Response, dict, None]) you can disable
# generating the response model from the type annotation with the path operation
# decorator parameter response_model=None.
#
# Read more: https://fastapi.tiangolo.com/tutorial/response-model/


@router.get('/{item_id}')
async def storage_retrieval_endpoint(
    request: FaRequest,
    item_id: str,
    background_tasks: BackgroundTasks,
) -> Any:
    """ Get authorized menu options """
    gs_request, other_params = get_default_fa_request()
    router.set_current_request(request, gs_request)
    other_params['response_type'] = other_params.get('response_type') or DEFAULT_DOWNLOAD_METHOD
    return storage_retieval_fa(request=gs_request, blueprint=router,
        item_id=item_id, other_params=other_params,
        background_tasks=background_tasks)


@router.get('/{item_id}/{response_type}')
async def storage_retrieval_with_response_type_endpoint(
    request: FaRequest,
    # It's optional to eventually report the error ASR-E1010
    item_id: str,
    response_type: str,
    background_tasks: BackgroundTasks,
) -> Any:
    """ Get authorized menu options """
    gs_request, other_params = get_default_fa_request()
    router.set_current_request(request, gs_request)
    other_params['response_type'] = response_type or DEFAULT_DOWNLOAD_METHOD
    return storage_retieval_fa(request=gs_request, blueprint=router,
        item_id=item_id, other_params=other_params,
        background_tasks=background_tasks)


def remove_temp_file(file_path: str) -> None:
    """ Remove the temp file """
    _ = DEBUG and log_debug(f"Removing temp file: {file_path}")
    os.remove(file_path)


def storage_retieval_fa(
    request: Request,
    blueprint: BlueprintOne,
    item_id: Union[str, None],
    other_params: Optional[Union[dict, None]] = None,
    background_tasks: Optional[BackgroundTasks] = None,
) -> Union[Response, FileResponse, StreamingResponse]:
    """
    Get S3 bucket content from encrypted item_id
    Args:
        request (Request): The request object.
        other_params (dict, optional): Other parameters. Defaults to None.
        item_id (str, optional): The item_id with the encrypted elements
            (bucket_name, separator and key). Defaults to None.
    Returns:
        Union[Response, StreamingResponse]: The object as streaming response
            or error response.
    """
    # settings = Config()
    other_params = other_params or {}

    if not other_params.get('response_type'):
        other_params['response_type'] = "fastapi"

    if other_params.get('response_type') in ["fastapi", "gs"]:
        other_params['mode'] = 'download'
    else:
        other_params['mode'] = 'get'

    resultset = storage_retieval(request=request, blueprint=blueprint,
        item_id=item_id, other_params=other_params)
    if resultset.get('error'):
        return return_resultset_jsonified_or_exception(
            resultset
        )

    if other_params.get('response_type') in ["fastapi", "gs"]:
        file_path = resultset['local_file_path']
        background_tasks.add_task(remove_temp_file, file_path=file_path)
        _ = DEBUG and log_debug(f"Temp file read | file_path: {file_path}")

    if other_params.get('response_type') == "gs":
        # Return the file content as GenericSuite way
        # (the one that worked for audio file and the ai_chatbot)
        _ = DEBUG and log_debug("Returning file content the Genericsuite way")
        return send_file_text_text(file_path)

    if other_params.get('response_type') == "fastapi":
        # Return the file content the standard FastAPI way
        # https://fastapi.tiangolo.com/advanced/custom-response/#fileresponse
        _ = DEBUG and log_debug("Returning file content as FileResponse")
        # return FileResponse(file_path, media_type=resultset['mime_type'])
        return FileResponse(file_path)

    if other_params.get('response_type') == "streaming":
        # Return the file content as a Streaming Response
        _ = DEBUG and log_debug("Returning file content as StreamingResponse")
        return StreamingResponse(io.BytesIO(resultset['content']),
            media_type=resultset['mime_type'])

    content_disposition_method = "inline"
    if other_params.get('response_type') == "attachment":
        content_disposition_method = "inline"

    # Return the file content as a normal Response
    headers = {
        'Content-Type': resultset['mime_type'],
        'Content-Disposition': f'{content_disposition_method}; filename=' \
            f'"{resultset["filename"]}"',
    }
    _ = DEBUG and log_debug("Returning file content as Response" +
        f' | headers: {headers}')
    return Response(
        body=resultset['content'],
        status_code=200,
        headers=headers
    )



"""
Chalice abstraction layer
"""
# from chalice.app import Request as ChaliceRequest, Response as ChaliceResponse
import os
import importlib

from genericsuite.chalicelib.util.blueprint_one import BlueprintOne as ChaliceBlueprintOne

DEBUG = False
FRAMEWORK_LOADED = False
FRAMEWORK = os.environ.get('CURRENT_FRAMEWORK', '').lower()
if FRAMEWORK == 'chalice':
    try:
        framework_module = importlib.import_module(FRAMEWORK)
        FRAMEWORK_LOADED = True
        if DEBUG:
            print(f'Chalicelib abstraction | framework_module: {framework_module}')


        class FrameworkClass(framework_module.Chalice):
            """
            Framkework class cloned from the selected framework super class.
            """


        class Request(framework_module.app.Request):
            """
            Request class cloned from the selected Request framework super class.
            This class is the one to be imported by the project modules
            """


        class Response(framework_module.app.Response):
            """
            Response class cloned from the selected Response framework super class.
            This class is the one to be imported by the project modules
            """


        class Blueprint(framework_module.app.Blueprint):
            """
            Blueprint class cloned from the selected Blueprint framework super class.
            This class is the one to be imported by the project modules
            """


        class BlueprintOne(ChaliceBlueprintOne):
            """
            Class to register a new route with optional schema validation and authorization.
            """

    except ImportError as err:
        raise ImportError(f"Unable to import '{FRAMEWORK}': {err}") from None



# send_email.py
# 2023-06-18 | CR

# https://realpython.com/python-send-email/

from os import environ
from os.path import basename
import smtplib
import ssl
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication
from email.utils import COMMASPACE, formatdate

from genericsuite.util.app_logger import log_debug, log_error

DEBUG = False


def send_email(sender_email, receiver_email, subject, text,
               html, files=None):
    """
    Send an Email
    """
    files = [] if not files else files
    smtp_server = environ.get('SMTP_SERVER')
    smtp_port = environ.get('SMTP_PORT')  # For starttls
    smtp_user = environ.get('SMTP_USER')
    smtp_password = environ.get('SMTP_PASSWORD')
    if sender_email is None or sender_email.strip() == '':
        sender_email = environ.get('SMTP_DEFAULT_SENDER')

    message = MIMEMultipart("alternative")
    message["Subject"] = subject
    message["From"] = sender_email
    message['To'] = COMMASPACE.join(receiver_email)
    message['Date'] = formatdate(localtime=True)

    # Turn these into plain/html MIMEText objects
    body_plain_text = MIMEText(text, "plain")
    if html:
        body_html = MIMEText(html, "html")

    # Add HTML/plain-text parts to MIMEMultipart message
    # The email client will try to render the last part first
    message.attach(body_plain_text)
    if html:
        message.attach(body_html)

    for file_hdl in files or []:
        with open(file_hdl, "rb") as fil:
            part = MIMEApplication(
                fil.read(),
                Name=basename(file_hdl)
            )
        # After the file is closed
        # part['Content-Disposition'] = 'attachment; filename="%s"' % basename(f)
        part['Content-Disposition'] = 'attachment; filename=' + \
                                      f'"{basename(file_hdl)}"'
        message.attach(part)

    if DEBUG:
        log_debug('SEND_EMAIL' +
                   f'\n | smtp_server: {smtp_server}' +
                   f'\n | smtp_port: {smtp_port}' +
                   f'\n | smtp_user: {smtp_user}' +
                   f'\n | smtp_password: {"*" * len(smtp_password)}' +
                   f'\n | sender_email: {sender_email}' +
                   f'\n | receiver_email: {receiver_email}' +
                   f'\n | subject: {subject}' +
                   f'\n | text: {text}' +
                   f'\n | html: {html}' +
                   f'\n | file_path: {files}')

    # Create a secure SSL context
    context = ssl.create_default_context()
    # Try to log in to smtp server and send email
    try:
        smtp = smtplib.SMTP(smtp_server, smtp_port)
        # smtp.ehlo()  # Can be omitted
        smtp.starttls(context=context)  # Secure the connection
        # smtp.ehlo()  # Can be omitted
        smtp.login(smtp_user, smtp_password)
    except Exception as err:
        # Print any error messages to stdout
        log_error(f'Send_Email ERROR (preparing phase): {err}')
        return False
    try:
        smtp.sendmail(sender_email, receiver_email, message.as_string())
    except Exception as err:
        # Print any error messages to stdout
        log_error(f'Send_Email ERROR (sending phase): {err}')
    finally:
        smtp.close()
    return True



"""
Configuration from the Database
"""
from typing import Any, Union, Optional
import os
import json

from genericsuite.util.app_context import (
    AppContext,
    ParamsFile,
    delete_params_file,
    PARAMS_FILE_ENABLED,
    PARAMS_FILE_GENERAL_FILENAME,
    NON_AUTH_REQUEST_USER_ID
)
from genericsuite.util.app_logger import log_debug, log_error
from genericsuite.util.generic_db_middleware import (
    fetch_all_from_db,
)
from genericsuite.util.jwt import AuthorizedRequest
from genericsuite.util.utilities import get_default_resultset


DEBUG = False
USE_DB_PARAMS_DEFAULT = os.environ.get('USE_DB_PARAMS_DEFAULT', "1")
# USE_DB_PARAMS_DEFAULT = "0"  # Usefull for slow local dev environments


def get_general_config(app_context: AppContext) -> dict:
    """
    Get all general parameters.
    """
    if DEBUG:
        log_debug('GGC-1) get_general_config')
    if os.environ.get("USE_DB_PARAMS", USE_DB_PARAMS_DEFAULT) != "1":
        return get_default_resultset()
    resultset = fetch_all_from_db(
        app_context=app_context,
        json_file='general_config',
        like_query_params={"active": "1"}
    )
    if not resultset["error"]:
        resultset["resultset"] = {
            r["config_name"]: r["config_value"]
            for r in json.loads(resultset["resultset"])
            if r["config_name"] not in
            # These variables cannot be loaded from database because
            # they are harmful if taken from source different
            # than enviroment variables.
            [
                "DB_CONFIG",
                "DB_ENGINE",
                "DEBUG",
                "APP_NAME",
                "APP_VERSION",
                "STAGE",
                "SECRET_KEY",
                "APP_SECRET_KEY",
                "APP_SUPERADMIN_EMAIL",
                "GIT_SUBMODULE_LOCAL_PATH",
                "CORS_ORIGIN",
                "HEADER_TOKEN_ENTRY_NAME",
                "USE_DB_PARAMS",
            ]
        }
    if DEBUG:
        log_debug('GGC-2) get_general_config |' +
                  f' resultset: {resultset}')
    return resultset


def get_users_config(app_context: AppContext) -> dict:
    """
    Get all user's parameters.
    """
    resultset = get_default_resultset()
    user_data = app_context.get_user_data()
    resultset["resultset"] = {r["config_name"]: r["config_value"]
                              for r in user_data.get("users_config", [])}
    if DEBUG:
        log_debug('GUC-2) get_users_config |' +
                  f' resultset: {resultset}')
    return resultset


def get_config_from_db_raw(app_context: AppContext) -> dict:
    """
    Get all dynamic parameters (general and user's).
    """
    if DEBUG:
        log_debug('GCFDR-1) get_config_from_db_raw')
    resultset = get_default_resultset()
    # Get general config from db
    config_from_db = get_general_config(app_context)
    if config_from_db["error"]:
        return config_from_db
    resultset['resultset'] = dict(config_from_db['resultset'].items())
    # Get user's config from db
    config_from_db = get_users_config(app_context)
    if config_from_db["error"]:
        return config_from_db
    resultset['resultset'].update(dict(config_from_db['resultset'].items()))
    if DEBUG:
        log_debug('GCFDR-2) get_config_from_db_raw |' +
                  f' resultset: {resultset}')
    return resultset


def get_all_params(app_context: AppContext):
    """
    Get all dynamic parameters (general and user's).
    First try from the json cache files, if not found get it from the
    database.
    """
    if PARAMS_FILE_ENABLED != '1':
        return get_config_from_db_raw(app_context)

    user_id = app_context.get_user_id()
    pfc = ParamsFile(user_id)

    # Try general params from the json file
    params = get_default_resultset()
    filename = pfc.get_params_file_path(PARAMS_FILE_GENERAL_FILENAME)
    load_result = pfc.load_params_file(filename)
    if load_result["found"]:  # and load_result['resultset']:
        params['resultset'].update(load_result['resultset'])
        _ = DEBUG and log_debug(
            'GCFD-4) app_context_and_set_env |' +
            ' General parameters loaded from file:' +
            f' {load_result["resultset"]}')
    else:
        # Get general params from json
        load_result = get_general_config(app_context)
        if load_result["error"]:
            return load_result
        params['resultset'].update(load_result['resultset'])
        # Save general params to json
        pfc.save_params_file(filename, load_result['resultset'])

    # Try user's config from json file
    if user_id != NON_AUTH_REQUEST_USER_ID:
        filename = pfc.get_params_filename()
        load_result = pfc.load_params_file(filename)
        if load_result["found"]:  # and load_result['resultset']:
            params['resultset'].update(
                {r["config_name"]: r["config_value"]
                    for r in load_result['resultset'].get("users_config", [])}
            )
            _ = DEBUG and log_debug(
                'GCFD-5) app_context_and_set_env |' +
                ' User\'s parameters loaded from file:' +
                f' {load_result["resultset"].get("users_config", [])}')
        else:
            # Get user's config from db
            load_result = get_users_config(app_context)
            if load_result["error"]:
                return load_result
            params['resultset'].update(dict(load_result['resultset'].items()))
            # Does not save the json file because it's a job for AppContex...
    return params


def app_context_and_set_env(request: AuthorizedRequest, blueprint: Any
                            ) -> AppContext:
    """
    Set the Appcontext and get all the parameters
    (general and user's) to dynamic set environment variables
    configured from the database.

    Args:
        request (AuthorizedRequest): the request object

    Returns:
        AppContext: the application context object, with
        the request object, user ID, user's data and the
        other object to expapnd the session dat.
    """
    app_context = AppContext()
    app_context.set_context_from_blueprint(blueprint=blueprint,
                                           request=request)
    if app_context.has_error():
        log_error('GCFD-0) app_context_and_set_env ERROR:'
                  f' {app_context.get_error()}')
        return app_context
    _ = DEBUG and \
        log_debug('GCFD-1) app_context_and_set_env')
    # Get all the parameters (general and user's) from dynamic set (database)
    params = get_all_params(app_context=app_context)
    if params["error"]:
        log_debug('GCFD-3) ERROR: app_context_and_set_env |' +
                  f' params: {params}')
        app_context.set_error(params["error_message"])
        return app_context
    for key, value in params['resultset'].items():
        # Set the environmet variable in the app_context
        # Previously it was "os.environ[key] = value" but it
        # carries a lot of issues...
        app_context.set_env_var(var_name=key, value=value)
    _ = DEBUG and \
        log_debug('GCFD-2) app_context_and_set_env |' +
                  f' Parameters set as os.environ(): {params["resultset"]}')
    return app_context


def set_init_custom_data(data: Optional[Union[dict, None]] = None):
    """
    Sets the custom data for the FastAPI/Flask/Chalice App.
    """
    result = dict(data.items()) if data else {}
    # Standard GenericDbHelper specific functions registry
    result['delete_params_file'] = delete_params_file
    _ = DEBUG and log_debug(f"//// Custom data: {result}")
    return result



# from chalice.app import (
#     UnauthorizedError,
#     BadRequestError,
#     ChaliceViewError,
#     ForbiddenError,
# )


# class GoAwayError(ChaliceViewError):
#     STATUS_CODE = 410


# class OverrideErrMsgMixin(Exception):

#     def __init__(self, msg: str = ''):  # noqa
#         self.msg = msg


# class QueryNotFound(GoAwayError, OverrideErrMsgMixin):
#     """Not found"""


# class QueryForbidden(ForbiddenError, OverrideErrMsgMixin):
#     """Query that field ss not allowed"""


# class QueryUnauthorized(UnauthorizedError, OverrideErrMsgMixin):
#     """Not authenticated """


# class QueryError(BadRequestError, OverrideErrMsgMixin):
#     """Generic Query Validation Failure"""


# class InternalError(ChaliceViewError, OverrideErrMsgMixin):
#     """Generic internal server error"""



"""
DbAbstractor: Database abstraction layer for MongoDb and DynamoDb
"""
from __future__ import annotations
from decimal import Decimal
import json

# from flask import current_app
# from Chalice import current_app

from bson.json_util import dumps, ObjectId
from werkzeug.local import LocalProxy

import pymongo
import boto3
import botocore

from genericsuite.util.app_logger import (
    log_debug,
    log_error,
    log_warning,
)
# from lib.models.dynamodb_table_structures \
#     import dynamodb_table_structures, \
#     DEFAULT_WRITE_CAPACITY_UNITS, DEFAULT_READ_CAPACITY_UNITS

from genericsuite.config.config import Config, is_local_service
from genericsuite.util.request_handler import RequestHandler

# ----------------------- Factory Methods -----------------------


DEBUG = False

DEFAULT_WRITE_CAPACITY_UNITS = 1
DEFAULT_READ_CAPACITY_UNITS = 1


class ObjectFactory:
    """
    Factory class for creating the different database objects.
    """
    def __init__(self):
        self._builders = {}

    def register_builder(self, key, builder):
        """
        Register the object (builder) for the given key (database).

        Args:
            key (str): The key for the database object.
            builder (DbAbstract): The database object (builder).

        Returns:
            None
        """
        self._builders[key] = builder

    def create(self, key, **kwargs):
        """
        Returns the given builder (database object) for the given key.

        Args:
            key (str): The key for the database object.
            **kwargs (dict): The keyword arguments for the database object.

        Returns:
            DbAbstract: The database object (builder).
        """
        builder = self._builders.get(key)
        if not builder:
            raise ValueError(key)
        return builder(**kwargs)


# ----------------------- Db Abstract -----------------------


class DbAbstract:
    """
    Database abstract super class
    """
    def __init__(self, app_config, **_ignored):
        self._app_config = app_config
        self._db = self.get_db_connection()
        # self.create_tables()

    def get_db(self):
        """
        Returns the database object.

        Returns:
            Object: The database object.
        """
        return self._db

    def get_db_connection(self):
        """
        Returns the database connection object.

        Returns:
            Object: The database connection object.
        """
        return {}

    def test_connection(self) -> str:
        """
        Test the database connection.

        Returns:
            str: The test result.
        """
        return dumps({})

    def list_collections(self, collection_name: str = None) -> list:
        """
        List the collections in the database.

        Returns:
            list: The list of collections.
        """
        return []

    def collection_stats(self, collection_name: str = None):
        """
        Get the collection statistics.

        Args:
            collection_name (str): the collection name. None means
            get all collections stats. Defaults to None.

        Returns:
            str: The collection(s) statistics.
        """
        return dumps(self.list_collections(collection_name))

    def create_tables(self) -> bool:
        """
        Create the tables in the database.
        """
        return True

    def table_exists(self, table_name: str) -> bool:
        """
        Check if the table exists in the database.

        Args:
            table_name (str): The name of the table.

        Returns:
            bool: True if the table exists, False otherwise.
        """
        return True


# ----------------------- MongoDb  -----------------------


class MongodbService(DbAbstract):
    """
    MongoDb service class
    """

    def get_db_connection(self):
        """
        Returns the database connection object.

        Returns:
            Object: The database connection object.
        """
        _ = DEBUG and \
            log_debug(
                "DB_ABSTRACTOR | MongodbService | get_db_connection" +
                # f"\n | DB_CONFIG: {self._app_config.DB_CONFIG}" +
                " | Starting...")
        client = pymongo.MongoClient(self._app_config.DB_CONFIG['mongodb_uri'])
        _ = DEBUG and \
            log_debug(
                "DB_ABSTRACTOR | MongodbService | get_db_connection" +
                f"\n | client: {client}" +
                "\n | DB Client OK...")
        db_connector = client.get_database(
            self._app_config.DB_CONFIG['mongodb_db_name'])
        _ = DEBUG and \
            log_debug(
                "DB_ABSTRACTOR | MongodbService | get_db_connection" +
                f"\n | db_connector: {db_connector}" +
                "\n | DB Connector OK...")
        return db_connector

    def test_connection(self) -> str:
        """
        Test the database connection.

        Returns:
            str: The test result.
        """
        return dumps(self._db.list_collection_names())

    def collection_stats(self, collection_name: str = None) -> str:
        """
        Get the collection statistics.

        Args:
            collection_name (str): the collection name. None means
            get all collections stats. Defaults to None.

        Returns:
            str: the MongoDb 'collstats' for all or the given collection
        """
        return dumps(self._db.command('collstats', collection_name))


class MongodbServiceBuilder(DbAbstract):
    """
    Builder class for MongoDb.
    """
    def __init__(self):
        self._instance = None

    def __call__(self, app_config, **_ignored):
        if not self._instance:
            self._instance = MongodbService(app_config)
        return self._instance


# ----------------------- DynamoDb  -----------------------


class DynamoDbUtilities:
    """
    DynamoDb Utilities class
    """
    def new_id(self):
        """
        Generate mongodb styled "_id"
        """
        # google: python generate unique _id like mongodb
        # https://www.geeksforgeeks.org/generating-random-ids-using-uuid-python/
        # id = uuid.uuid1()
        # uuid1() includes the used of MAC address of computer,
        # and hence can compromise the privacy,
        # even though it provides UNIQUENES.
        # return id.hex
        # RESULT: bson.errors.InvalidId: '48a8b1a021b611edbf5a0e4c731ac1c1'
        # is not a valid ObjectId, it must be a 12-byte input or a
        # 24-character hex string | vs 6302ded424b11a2032d7c562
        # SOLUTION: https://api.mongodb.com/python/1.11/api/bson/objectid.html
        return str(ObjectId())

    def id_addition(self, row):
        """
        Convert _id to be mongodb styled,
        for example to send it to the react frontend
        expecting it as a $oid
        """
        if 'id' in row:
            row['_id'] = ObjectId(row['id'])
        elif '_id' in row and isinstance(row['_id'], str):
            # row['id'] = row['_id']
            row['_id'] = ObjectId(row['_id'])
        return row

    def id_conversion(self, key_set):
        """
        To avoid error working internally with mongodb styled "_id"
        """
        if DEBUG:
            log_debug('**** id_conversion | key_set BEFORE: ' + str(key_set))
        if '_id' in key_set and not isinstance(key_set['_id'], str):
            key_set['_id'] = str(key_set['_id'])
        if DEBUG:
            log_debug('**** id_conversion | key_set AFTER: ' + str(key_set))
        return key_set

    def convert_floats_to_decimal(self, data):
        """
        To be used before sending dict to DynamoDb for inserts/updates
        """
        # return json.loads(json.dumps(item), parse_float=Decimal)
        #   --> TypeError: Object of type Decimal is not JSON serializable
        # return json.loads(json.dumps(item, default=float))
        #   --> Error updating Item [UO_ERR_020]: Float types are not
        #       supported. Use Decimal types instead.
        if isinstance(data, dict):
            return {k: self.convert_floats_to_decimal(v)
                    for k, v in data.items()}
        elif isinstance(data, list):
            return [self.convert_floats_to_decimal(item) for item in data]
        elif isinstance(data, float):
            return Decimal(str(data))
        else:
            return data

    def remove_decimal_types(self, item: dict, projection: dict = None):
        """
        To be used before sending responses to entity
        """
        if not projection:
            projection = {}
        if DEBUG:
            log_debug('====> REMOVE_DECIMAL_TYPES' +
                      f' | projection: {projection}' +
                      f' | item BEFORE: {item}')
        # Convert MongoDB _id Object to str
        item = self.id_conversion(item)
        # Convert Decimal to floats
        item = json.loads(json.dumps(item, default=float))
        # Convert _id to be mongodb styled
        item = self.id_addition(item)
        # Applying MongoDB like projection (replacing the use of DymanoDb's
        # ProjectionExpression)
        item = {k: v for k, v in item.items() if projection.get(k, 1) == 1}
        if DEBUG:
            log_debug('====> REMOVE_DECIMAL_TYPES | item AFTER: ' + str(item))
        return item

    def remove_decimal_types_list(self, items: list, projection: dict = None):
        """
        To be used before sending responses to entity as a list
        """
        if items is None:
            _ = DEBUG and \
                log_debug('remove_decimal_types_list | None result: []')
            return []
        return [self.remove_decimal_types(item, projection) for item in items]


class DynamoDbFindIterator(DynamoDbUtilities):
    """
    Dynamodb find iterator
    """
    def __init__(self, data_set):
        if DEBUG:
            log_debug(
                '>>--> DynamoDbFindIterator | __init__() | data_set: ' +
                str(data_set)
            )
        self._data_set = data_set
        self._skip = 0
        self._limit = None
        self._num = None

    def skip(self, skip):
        """
        Set the skip value
        """
        if DEBUG:
            log_debug('>>--> DynamoDbFindIterator | skip() | skip: ' +
                      str(skip))
        self._skip = skip
        return self

    def limit(self, limit):
        """
        Set the limit value
        """
        if DEBUG:
            log_debug(
                '>>--> DynamoDbFindIterator | limit() | limit: ' + str(limit)
            )
        self._limit = limit
        return self

    def __iter__(self):
        self._num = self._skip
        if DEBUG:
            log_debug(
                '>>--> DynamoDbFindIterator | __iter__() | self.num: ' +
                str(self._num)
            )
        return self

    def __next__(self):
        if DEBUG:
            log_debug(
                '>>--> DynamoDbFindIterator | __next__() | self.num: ' +
                str(self._num) + ' | len(self._data_set): ' +
                str(len(self._data_set))
            )
        if isinstance(self._data_set, dict):
            if self._num > 0:
                raise StopIteration
            self._num += 1
            return self.remove_decimal_types(
                self.id_addition(self._data_set)
            )
        if (not self._limit or self._num <= self._limit) and \
           self._data_set and self._num < len(self._data_set):
            # _result = self.convert_floats_to_decimal(
            #   self.id_addition(self._data_set[self._num])
            # )
            _result = self.remove_decimal_types(
                self.id_addition(self._data_set[self._num])
            )
            if DEBUG:
                log_debug(
                    '>>--> DynamoDbFindIterator | __next__() | _result: ' +
                    str(_result)
                )
            # _result = self.id_addition(self._data_set[self._num])
            self._num += 1
            return _result
        raise StopIteration

    def sort(self, column: str, direction: str):
        """
        Sort the data set
        """
        if isinstance(self._data_set, dict):
            return self
        if self._data_set is None:
            self._data_set = []
        else:
            self._data_set.sort(
                key=lambda data_set: data_set.get(column),
                reverse=(direction != 'asc'))
        return self


class DynamoDbTableAbstract(DynamoDbUtilities):
    """
    DynamoDb Table Abstract class
    """
    def __init__(self, item_structure, db_conection):
        if DEBUG:
            log_debug(
                '>>--> DynamoDbTableAbstract | __init__ | item_structure: ' +
                str(item_structure)
            )
        self._prefix = item_structure['prefix']
        self._table_name = item_structure['TableName']
        self._key_schema = None
        self._attribute_definitions = None
        self._global_secondary_indexes = None
        self._db_conection = db_conection
        self.inserted_id = None
        self.modified_count = None
        self.deleted_count = None

    def get_table_definitions(self):
        """
        The first time the table is used, get the table definitions
        """
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb/client/describe_table.html
        response = self._db_conection.meta.client.describe_table(
            TableName=self.get_table_name()
        )
        _ = DEBUG and log_debug(f'|||---> get_table_definitions: {response}')
        item_structure = response['Table']
        self._key_schema = item_structure['KeySchema']
        self._attribute_definitions = item_structure['AttributeDefinitions']
        self._global_secondary_indexes = \
            item_structure.get('GlobalSecondaryIndexes', [])

    def get_table_name(self):
        """
        Get the table name, adding the prefix
        """
        table_name = f'{self._prefix}{self._table_name}'
        _ = DEBUG and log_debug(f'|||---> get_table_name: {table_name}')
        return table_name

    def get_key_schema(self):
        """
        Get the table key schema
        """
        if self._key_schema is None:
            self.get_table_definitions()
        return self._key_schema

    def get_attribute_definitions(self):
        """
        Get the table attribute definitions
        """
        if self._attribute_definitions is None:
            self.get_table_definitions()
        return self._attribute_definitions

    def get_global_secondary_indexes(self):
        """
        Get the table global secondary indexes
        """
        if self._global_secondary_indexes is None:
            self.get_table_definitions()
        return self._global_secondary_indexes

    def element_name(self, element_name):
        """
        Get the element's attribute name 'AttributeName'
        """
        if DEBUG:
            log_debug('|||---> element_name: ' + str(element_name))
        # return element_name['AttributeName'] \
        #   if element_name['AttributeName'] != '_id' else 'id'
        return element_name['AttributeName']

    def get_conditions(self, key, value):
        """
        Get the conditions for the DynamoDB query.
        Handles the mongoDB conditions:
            key=value, $regex, $eq, $ne, $gt, $gte, $lt, $lte, $in, $nin

        Args:
            key (str): the key to search for
            value (any): the value to search for. If value is a dict,
                it contains the conditions to apply. E.g.
                {'$regex': '.*any_value.*', '$options': 'si'}
                {'$eq': 'any_value'}
                {'$ne': 'any_value'}
                {'$gt': 'any_value', '$lt': 'any_value'}
                {'$gte': 'any_value', '$lte': 'any_value'}
                {'$in': ['any_value', 'any_value']}
                {'$nin': ['any_value', 'any_value']}

        Returns:
            list: a list of conditions to apply to the query, each one is a
                dict with the following keys:
                    'attr': the attribute name
                    'key': the key name for the condition
                    'comp_oper': the comparison operator
                    'value': the value to compare against
                    'function': the function to apply (if any)
        """
        conditions = []
        if not isinstance(value, dict):
            conditions.append({
                'attr': key,
                'key': key,
                'comp_oper': '=',
                'value': value,
            })
        else:
            # https://www.mongodb.com/docs/manual/reference/operator/query-comparison/
            # https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/Query.FilterExpression.html
            # https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/Query.KeyConditionExpressions.html
            # https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/Expressions.OperatorsAndFunctions.html#Expressions.OperatorsAndFunctions.Syntax
            if '$regex' in value:
                # It's a regex. E.g. 'attribute_name':
                #     {'$regex': '.*any_value.*', '$options': 'si'}
                conditions.append({
                    'attr': key,
                    'key': f'{key}Rg',
                    'value': value['$regex'].replace('.*', ''),
                    'function': f'contains ({key}, :{key}Rg)',
                })
            if '$eq' in value:
                conditions.append({
                    'attr': key,
                    'key': f'{key}Eq',
                    'comp_oper': '=',
                    'value': value['$eq'],
                })
            if '$ne' in value:
                conditions.append({
                    'attr': key,
                    'key': f'{key}Ne',
                    'comp_oper': '<>',
                    'value': value['$ne'],
                })
            if '$gt' in value:
                conditions.append({
                    'attr': key,
                    'key': f'{key}Gt',
                    'comp_oper': '>',
                    'value': value['$gt'],
                })
            if '$gte' in value:
                conditions.append({
                    'attr': key,
                    'key': f'{key}Gte',
                    'comp_oper': '>=',
                    'value': value['$gt'],
                })
            if '$lt' in value:
                conditions.append({
                    'attr': key,
                    'key': f'{key}Lt',
                    'comp_oper': '<',
                    'value': value['$lt'],
                })
            if '$lte' in value:
                conditions.append({
                    'attr': key,
                    'key': f'{key}Lte',
                    'comp_oper': '<=',
                    'value': value['$lte'],
                })
            if '$in' in value or '$nin' in value:
                # https://stackoverflow.com/questions/40283653/how-to-use-in-statement-in-filterexpression-using-array-dynamodb
                cond_key = '$in' if '$in' in value else '$nin'
                i = 0
                values = {}
                key_list = []
                for v in value[cond_key]:
                    i += 1
                    key = f'{key}In{i}'
                    key_list.append(f':{key}')
                    values[key] = v
                    # values[key] = f'"{v}"' if isinstance(v, str) else v
                conditions.append({
                    'value': values,
                    'function':
                        ('NOT ' if '$nin' in value else '') +
                        f"{key} IN ({','.join(key_list)})",
                })
        return conditions

    def get_cond_exp_val(self, item):
        """
        Get the condition expression values for a DynamoDB operation.

        Args:
            item (dict): A dictionary representing the item to search for.

        Returns:
            tuple: A tuple containing the condition values dictionary, the
                   condition expression string and the attribute names.
        """
        condition_values = {}
        expresion_parts = []
        attr_names = {}
        for key in item.keys():
            conditions = self.get_conditions(key, item[key])
            group_char = ''
            inner_separator = ''
            for condition in conditions:
                attr = condition.get('attr')
                key = condition.get('key')
                comp_oper = condition.get('comp_oper')
                value = condition.get('value')
                multiple = condition.get('multiple')
                function = condition.get('function')
                if multiple is None:
                    group_char = ''
                    open_group = ''
                    close_group = ''
                elif multiple:
                    group_char = '(' if group_char == '' else ')'
                    open_group = '(' if group_char == '(' else ''
                    close_group = ')' if group_char == ')' else ''
            # Populates the condition values (to be used as the
            # ExpressionAttributeValues parameter)
            if value:
                if not isinstance(value, dict):
                    value = {key: value}
                for key_name, key_value in value.items():
                    condition_values = condition_values | \
                        {':' + key_name: key_value}
            # Populates the condition expressions (to be used as the
            # KeyConditionExpression or FilterExpression parameter)
            if function is None:
                # It's a condition
                expresion_parts.append(
                    f'{inner_separator}' +
                    f'{open_group}#{attr}_ {comp_oper} :{key}{close_group}')
                # Set an attribute name to be passed in the
                # ExpressionAttributeNames parameter calling the scan, query
                # or updat_item DynamoDB operations, to avoid errors like:
                # "Invalid UpdateExpression: Attribute name is a reserved
                # keyword; reserved keyword: language"
                attr_names[f'#{attr}_'] = attr
            else:
                # It's a function call
                expresion_parts.append(
                    f'{inner_separator}' +
                    f'{open_group}{function}{close_group}')
            # It has more than one condition
            if multiple is not None:
                inner_separator = condition.get('separator', 'AND') + ' '
        return condition_values, expresion_parts, attr_names

    def get_condition_expresion_values(self, data_list, separator=' AND '):
        """
        Generate the condition expression values for a DynamoDB operation.

        Args:
            data_list (list): A list of dictionaries, where each dictionary
                              represents one item's attributes for the
                              condition.
            separator (str): The separator to use between conditions.
            Defaults to ' AND '.

        Returns:
            tuple: A tuple containing the condition values dictionary, the
                   condition expression string and the attribute names.
        """
        condition_values = {}
        expresion_parts = []
        attr_names = {}
        _ = DEBUG and log_debug(
            '|||---> get_condition_expresion_values | ' +
            f'data_list: {data_list} | separator: {separator}')
        for item in data_list:
            _ = DEBUG and log_debug('|||---> get_condition_expresion_values' +
                                    f' | item: {item}')
            if '$and' in item or '$or' in item:
                separator = '$or' if '$or' in item else '$and'
                for subitem in item[separator]:
                    _ = DEBUG and log_debug(
                        '|||---> get_condition_expresion_values' +
                        f' | subitem: {subitem}')
                    condition_values, expresion_parts, attr_names = \
                        self.get_cond_exp_val(subitem)
            else:
                condition_values, expresion_parts, attr_names = \
                    self.get_cond_exp_val(item)
        separator = ' OR ' if '$or' in separator else ' AND ' \
            if '$and' in separator else separator
        condition_expresion = separator.join(expresion_parts)
        return condition_values, condition_expresion, attr_names

    def get_primary_keys(self, query_params):
        """
        Look for keys in partition/sort key
        """
        keys = list(filter(lambda key:
                    query_params.get(self.element_name(key)) is not None,
                    self.get_key_schema()))
        keys = list(map(lambda key: {self.element_name(key):
                    query_params.get(self.element_name(key))}, keys))
        return keys[0] if len(keys) > 0 else None

    def get_global_secondary_indexes_keys(self, query_params):
        """
        Look for keys in global secondary indexes
        """
        reduced_indexes = [
            {
                'name': global_index["IndexName"],
                'keys':
                    list(
                        map(
                            lambda key: self.element_name(key),
                            global_index["KeySchema"]
                        )
                    )
            } for global_index in self.get_global_secondary_indexes()
        ]

        query_keys = list(query_params.keys())

        index_item = list(
            filter(
                lambda index: set(query_keys).issubset(index['keys']),
                reduced_indexes
            )
        )

        keys = None
        index_name = None
        if index_item:
            index_name = index_item[0]['name']
            keys = list(
                map(
                    lambda key: {key: query_params[key]}, index_item[0]['keys']
                )
            )

        if DEBUG:
            log_debug(
                '|-|--> get_global_secondary_indexes_keys | keys: ' +
                str(keys) +
                ' | reduced_indexes: ' + str(reduced_indexes) +
                ' | index_name: ' +
                str(index_name) + ' | query_keys: ' + str(query_keys))
        return keys, index_name

    def generic_query(self, query_params: dict, projection: dict = None,
                      query_type: str = 'find', select: str = None) -> list:
        """
        Perform a query on the DynamoDB table from MongoDb style query params.

        Args:
            query_params (dict): The filter for the MongoDb query.
            projection (dict): The projection for the query.
                Uses the ProjectionExpression attribute.
                Defaults to an empty dictionary.
            query_type (str): query type: 'find' get one or more items,
                'find_one': get one item.
            select (str): The select for the query.
                Possible values: 'ALL_ATTRIBUTES' | 'ALL_PROJECTED_ATTRIBUTES'
                | 'SPECIFIC_ATTRIBUTES' | 'COUNT'.
                Defaults to None, meaning ALL_ATTRIBUTES.

        Returns:
            list: The result of the query.
        """
        if not projection:
            projection = {}
        select = select or 'ALL_ATTRIBUTES'
        projection_expression = None
        if select == 'SPECIFIC_ATTRIBUTES':
            # Assumes the specific attributes to retrieve are in the
            # projection parameter with a value of 1
            projection_expression = (",".join([k for k, v
                                     in projection.items() if v == 1]))
        select = select.upper()
        if DEBUG:
            log_debug(
                f'>>--> generic_query() | table: {self.get_table_name()}' +
                f' | query_params: {query_params}' +
                f' | projection: {projection}' +
                f' | query_type: {query_type}' +
                f' | select: {select}')

        table = self._db_conection.Table(self.get_table_name())

        if not query_params or len(query_params) == 0:
            response = table.scan()
            if select == "COUNT":
                count = len(response.get('Items', []))
                _ = DEBUG and \
                    log_debug(f'generic_query | COUNT 4: {count}')
                return count
            _ = DEBUG and \
                log_debug('generic_query | response.get(Items):' +
                          f' {response.get("Items")}')
            return self.remove_decimal_types_list(
                response.get('Items', []), projection)

        top_and_or = '$and' in query_params or '$or' in query_params
        keys = None

        if not top_and_or:
            query_params = self.id_conversion(query_params)
            keys = self.get_primary_keys(query_params)
            if keys:
                # Get only one item
                _ = DEBUG and log_debug('generic_query | ===> Keys found: ' +
                                        str(keys))
                # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb/table/get_item.html
                params = {
                    'Key': keys,
                }
                if projection_expression:
                    params['ProjectionExpression'] = projection_expression
                response = table.get_item(**params)
                _ = DEBUG and log_debug(response)
                if select == "COUNT":
                    count = 1 if response and response.get('Item') else 0
                    _ = DEBUG and \
                        log_debug(f'generic_query | COUNT 1: {count}')
                    return count
                return self.remove_decimal_types(response.get('Item', {}),
                                                 projection)
            keys, index_name = \
                self.get_global_secondary_indexes_keys(query_params)

        if not keys:
            condition_values, condition_expresion, attr_names = \
                self.get_condition_expresion_values([query_params])
            if DEBUG:
                log_debug(
                    'generic_query | ' +
                    'No keys found... perform fullscan | condition_values: ' +
                    str(condition_values) + ' | condition_expresion: ' +
                    str(condition_expresion)
                )
            # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb/table/scan.html
            params = {
                'ExpressionAttributeValues': condition_values,
                'Select': select,
            }
            if condition_expresion:
                params['FilterExpression'] = condition_expresion
            if projection_expression:
                params['ProjectionExpression'] = projection_expression
            if attr_names:
                params['ExpressionAttributeNames'] = attr_names
            response = table.scan(**params)
            if query_type == 'find_one':
                # Get only one item
                if select == "COUNT":
                    count = 1 if response and response.get('Items') and \
                        len(response.get('Items')) > 0 else 0
                    _ = DEBUG and \
                        log_debug(f'generic_query | COUNT 2: {count}')
                    return count
                return self.remove_decimal_types(
                    response.get('Items')[0] if response and
                    response.get('Items') and
                    len(response.get('Items')) > 0 else {},
                    projection)
            # Get more than one item
            if select == "COUNT":
                count = len(response.get('Items', []))
                _ = DEBUG and \
                    log_debug(f'generic_query | COUNT 3: {count}')
                return count
            return self.remove_decimal_types_list(
                response.get('Items', []), projection)

        if DEBUG:
            log_debug(f'generic_query | ===> secondary Keys found: {keys}')
        condition_values, condition_expresion, attr_names = \
            self.get_condition_expresion_values(keys)
        if DEBUG:
            log_debug(
                'generic_query | ' +
                '===> secondary Keys pre query elements | condition_values: ' +
                str(condition_values) + ' | condition_expresion:' +
                str(condition_expresion)
            )
        params = {
            'IndexName': index_name,
            'ExpressionAttributeValues': condition_values,
            'KeyConditionExpression': condition_expresion,
            'Select': select,
        }
        if condition_expresion:
            params['FilterExpression'] = condition_expresion
        if projection_expression:
            params['ProjectionExpression'] = projection_expression
        if attr_names:
            params['ExpressionAttributeNames'] = attr_names
        response = table.query(**params)
        if DEBUG:
            log_debug(response)
        if query_type == 'find_one':
            # Get only one item
            return self.remove_decimal_types(
                response.get('Items', [])[0], projection)
        # Get more than one item
        return self.remove_decimal_types_list(response.get('Items', []),
                                              projection)

    def find(self, query_params, projection=None):
        """
        Translate MongoDb 'find' to DynamoDb 'query' and returns the iterator

        The MongoDb call style:

        resultset['resultset'] = db.users.find({}, projection).skip(
            int(skip)
        ).limit(int(limit))
        """
        if not projection:
            projection = {}
        if DEBUG:
            log_debug(
                '>>--> find() | table: ' + self.get_table_name() +
                ' | query_params: ' + str(query_params) + ' | projection: ' +
                str(projection)
            )
        return DynamoDbFindIterator(
            self.generic_query(query_params, projection, query_type='find'))

    def find_one(self, query_params, projection=None):
        """
        Translate MongoDb 'find_one' to DynamoDb 'query'' and returns the
        result.

        The MongoDb call style:

        resultset['resultset'] = db.users.find_one({'_id': id}, projection)
        """
        if not projection:
            projection = {}
        if DEBUG:
            log_debug(
                '>>--> find_one() | table: ' + self.get_table_name() +
                ' | query_params: ' + str(query_params) + ' | projection: ' +
                str(projection)
            )
        return self.generic_query(query_params, projection,
                                  query_type='find_one')

    def insert_one(self, new_item):
        """
        Translate MongoDb 'insert_one' to DynamoDb 'put_item' and returns
        the result.

        The MongoDb call style:

        resultset['resultset']['_id'] = str(
            db.users.insert_one(json).inserted_id
        )
        """
        table = self._db_conection.Table(self.get_table_name())
        self.inserted_id = None
        new_item['_id'] = self.new_id()
        new_item = self.convert_floats_to_decimal(new_item)
        _ = DEBUG and log_debug(f'insert_one | new_item: {new_item}')
        try:
            result = table.put_item(Item=new_item)
            self.inserted_id = new_item['_id']
            if DEBUG:
                log_debug(
                    '>>--> RESULT insert_one() | table: ' +
                    self.get_table_name() +
                    ' | new_item: ' + str(new_item) + ' | self.inserted_id: ' +
                    str(self.inserted_id) + ' | result: ' + str(result)
                )
        except botocore.exceptions.ClientError as err:
            log_error(
                'insert_one: Error creating Item [IO_ERR_010]: ' + str(err))
            raise err
        except Exception as err:
            log_error(
                'insert_one: Error creating Item [IO_ERR_020]: ' + str(err))
            raise err
        return self

    def replace_one(self, key_set, update_set_original):
        """
        Translate MongoDb 'replace_one' to DynamoDb 'update_item' and returns
        the result.

        The MongoDb call style:
            (see update_one)
        """
        # Replace what is was in the item
        return self.update_one(key_set, update_set_original)
        # return self.update_one(key_set, {'$set': update_set_original})

    def update_one(self, key_set, update_set_original):
        """
        Translate MongoDb 'update_one' to DynamoDb 'update_item' and returns
        the result.

        The MongoDb call style:

        Case 1: $set
        # Update the item, preserving the attributes not present in
        # the updated_record

        resultset['resultset']['rows_affected'] = str(
          db.users.update_one(
             {'_id': ObjectId(record['_id'])},
             {'$set': updated_record}
          ).modified_count
        )

        Case 2: $addToSet
        # Add a new element to an array in the item

        resultset['resultset']['rows_affected'] = str(
          db.users.update_one(
              {'_id': ObjectId(json[parent_key_field])},
              {'$addToSet': {array_field: json[array_field]}}
          ).modified_count
        )

        Case 3: $pull
        # Remove an existing element from an array in the item

        resultset['resultset']['rows_affected'] = str(
          db.users.update_one(
              {'_id': ObjectId(json[parent_key_field])},
              {'$pull': {
                  array_field: {
                      array_key_field: \
                        json[array_field_in_json][array_key_field]
                  }
              }}
          ).modified_count
        )

        Case 4: none of the above
        # Replace what is was in the item

        resultset['resultset']['rows_affected'] = str(
          db.users.update_one(
             {'_id': ObjectId(record['_id'])},
             updated_record
          ).modified_count
        )


        """
        if DEBUG:
            log_debug(
                '>>--> update_one() | table: ' + self.get_table_name() +
                f' | key_set: {key_set}')
        table = self._db_conection.Table(self.get_table_name())
        key_set = self.convert_floats_to_decimal(self.id_conversion(key_set))
        self.modified_count = None
        keys = self.get_primary_keys(key_set)
        _ = DEBUG and log_debug(f'>>--> update_one() | keys: {keys}')
        if not keys:
            log_warning('update_one: No partition keys found [UO_ERR_010]')
            return False

        result = None
        if '$set' in update_set_original or \
           '$addToSet' in update_set_original or \
           '$pull' in update_set_original:
            try:
                result = table.get_item(Key=keys)
                _ = DEBUG and log_debug('>>--> update_one() | result:' +
                                        f' {result}')
            except Exception as err:
                log_error('update_one: Error getting existing Item' +
                          f' [UO_ERR_030]: {str(err)}')
                return False
            if not result.get('Item'):
                log_error('update_one: Item not found [UO_ERR_040]')
                return False

        _ = DEBUG and log_debug('>>--> update_one() | update_set_original:' +
                                f' {update_set_original}')
        if '$set' in update_set_original:
            # Update the item, preserving the attributes not present in
            # the update_set_original
            update_set = result['Item'].update(update_set_original['$set'])
        elif '$addToSet' in update_set_original:
            # Add a new element to an array in the item
            array_field, array_value = next(
                iter(update_set_original['$addToSet'].items())
            )
            if array_field in result['Item']:
                result['Item'][array_field].append(array_value)
            else:
                result['Item'][array_field] = [array_value]
            update_set = result['Item']
        elif '$pull' in update_set_original:
            # Remove an existing element from an array in the item
            array_field, array_value = next(
                iter(update_set_original['$pull'].items())
            )
            if array_field in result['Item']:
                result['Item'][array_field].remove(array_value)
            update_set = result['Item']
        else:
            # Replace what is was in the item
            update_set = update_set_original

        # To avoid the error:
        # "Float types are not supported. Use Decimal types instead.""
        update_set = self.convert_floats_to_decimal(update_set)

        # Don't include the PK / SK (Primary Key / Sort Key)
        update_set = {k: v for k, v in update_set.items() if k not in keys}

        # Prepare update expresions and values
        expression_attribute_values, update_expression, attr_names = \
            self.get_condition_expresion_values([update_set], ', ')

        _ = DEBUG and log_debug(
            '>>--> update_one()' +
            f'\n| update_set: {update_set}' +
            f'\n| expression_attribute_values: {expression_attribute_values}' +
            f'\n| update_expression: {update_expression}'
        )

        try:
            if DEBUG:
                log_debug(
                    '>>--> BEFORE update_one()' +
                    f' | table: {self.get_table_name()}' +
                    f'\n| update_set: {update_set}' +
                    f'\n| keys: {keys}' +
                    f'\n| self.modified_count: {self.modified_count}' +
                    '\n| expression_attribute_values: ' +
                    f'{expression_attribute_values}' +
                    f'\n| update_expression: {update_expression}'
                )
            params = {
                'Key': keys,
                'UpdateExpression': "SET " + update_expression,
                'ExpressionAttributeValues': expression_attribute_values,
                'ReturnValues': "UPDATED_NEW"
            }
            if attr_names:
                params['ExpressionAttributeNames'] = attr_names
            result = table.update_item(**params)
            self.modified_count = 1
            if DEBUG:
                log_debug(
                    '>>--> RESULT update_one() | table: ' +
                    self.get_table_name() +
                    ' | update_set: ' + str(update_set) + ' | keys: ' +
                    str(keys) +
                    ' | self.modified_count: ' + str(self.modified_count) +
                    ' | expression_attribute_values: ' +
                    str(expression_attribute_values) +
                    ' | update_expression: ' +
                    str(update_expression) + ' | result: ' + str(result)
                )
            return self
        except Exception as err:
            log_error(
                'update_one: Error updating Item [UO_ERR_020]: ' + str(err))
        return False

    def delete_one(self, key_set):
        """
        Translate MongoDb 'delete_one' to DynamoDb 'delete_item' and returns
        the result.

        The MongoDb call style:

        resultset['resultset']['rows_affected'] = str(db.users.delete_one(
          {'_id': ObjectId(user_id)}).deleted_count
        )
        """
        if DEBUG:
            log_debug(
                '>>--> delete_one() | table: ' + self.get_table_name() +
                ' | key_set: ' + str(key_set)
            )
        table = self._db_conection.Table(self.get_table_name())
        key_set = self.id_conversion(key_set)
        self.deleted_count = None
        keys = self.get_primary_keys(key_set)
        if not keys:
            log_warning('delete_one: No partition keys found [DO_ERR_010]')
            return False
        try:
            result = table.delete_item(Key=keys)
            self.deleted_count = 1
            if DEBUG:
                log_debug(
                    '>>--> RESULT delete_one() | table: ' +
                    self.get_table_name() +
                    ' | key_set: ' + str(key_set) + ' | keys: ' + str(keys) +
                    ' | self.deleted_count: ' + str(self.deleted_count) +
                    ' | result: ' + str(result)
                )
            return self
        except Exception as err:
            log_error(
                'delete_one: Error updating Item [DO_ERR_020]: ' + str(err)
            )
        return False

    def count_documents(self, filter):
        """
        Translate MongoDb 'count_documents' to DynamoDb 'scan' and returns
        the result.

        The MongoDb call style:

        resultset['resultset']['count'] = str(
            db.collection_name.count_documents(filter)
        )
        """
        return self.generic_query(query_params=filter, query_type='find',
                                  select="COUNT")


class DynamodbServiceSuper(DbAbstract, DynamoDbUtilities):
    """
    Dynamodb Service super class
    """
    def get_db_connection(self):
        """
        Get the DynamoDB connection
        """
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb.html
        self._db_params = {}
        if is_local_service():
            self._db_params['endpoint_url'] = \
                self._app_config.DB_CONFIG['mongodb_uri']
        self._db = boto3.resource('dynamodb', **self._db_params)
        self._prefix = self._app_config.DB_CONFIG['dynamdb_prefix']
        self.create_table_name_propeties()
        return self._db

    def create_table_name_propeties(self):
        """
        Create DynamoDb table name class propeties so tables can be retrieved
        as a subscript (like MongoDB tables) using the __getitem__() method.
        """
        item_list = self.list_collections(prefix=self._prefix)
        for item_name in item_list:
            item_props = {
                "TableName": item_name.replace(self._prefix, ''),
                "prefix": self._prefix,
            }
            if DEBUG:
                log_debug(
                    '||| create_table_name_propeties' +
                    f'\n>>--> Setting property: {item_name}' +
                    f' | item_props: {item_props}')
            setattr(
                self,
                item_name,
                DynamoDbTableAbstract(item_props, self._db)
            )

    def list_collection_names(self):
        """
        Returns a list with the collection (table) names
        """
        return map(lambda item_name: item_name, self.list_collections())

    def get_db(self):
        """
        Returns the database object.

        Returns:
            Object: The database object. For DynamoDb, it must returns this
                Class as a whole.
        """
        return self

    def test_connection(self):
        """
        Test the database connection. If the table list can be retrieved,
        the connection is OK.

        Returns:
            str: A JSON string with the table list.
        """
        return dumps(self.list_collection_names(prefix=self._prefix))

    def create_tables(self, dynamodb_table_structures: dict):
        """
        Create the tables in the DynamoDB database.
        """
        default_provisioned_throughput = {
            'ReadCapacityUnits': DEFAULT_READ_CAPACITY_UNITS,
            'WriteCapacityUnits': DEFAULT_WRITE_CAPACITY_UNITS
        }
        item_list = dynamodb_table_structures.keys()
        for item_name in item_list:
            item_props = dynamodb_table_structures.get('Table')
            if not item_props:
                continue
            item_name = item_props["TableName"]
            if DEBUG:
                log_debug('>>--> Creating Dynamodb Table: ' + item_name)
            if self.table_exists(item_name):
                continue
            # Create table in Dynamodb
            table = self._db.create_table(
                TableName=item_name,
                KeySchema=item_props['KeySchema'],
                AttributeDefinitions=item_props
                ['AttributeDefinitions'],
                ProvisionedThroughput=item_props.get(
                    'provisioned_throughput', default_provisioned_throughput
                ),
                GlobalSecondaryIndexes=item_props.get(
                    'GlobalSecondaryIndexes', []
                ),
            )
            # Wait until the table exists.
            table.wait_until_exists()
            # Print out some data about the table.
            if DEBUG:
                log_debug(
                    '>>--> Dynamodb Table: ' + item_name +
                    ' CREATED! - table.item_count: ' + str(table.item_count)
                )

        return True

    def table_exists(self, table_name: str) -> bool:
        """
        Check if the table exists in the database.

        Args:
            table_name (str): The name of the table.

        Returns:
            bool: True if the table exists, False otherwise.
        """
        try:
            self._db.Table(table_name).table_status
        except self._db.meta.client.exceptions.ResourceNotFoundException:
            return False
        return True


class DynamodbService(DynamodbServiceSuper):
    """
    Class for DynamodbService.
    """
    def list_collections(self, collection_name: str = None,
                         prefix: str = None):
        """
        List all or filtered tables in the DynamoDB database.

        Args:
            collection_name (str): The name of one table to filter by.
            prefix (str): The prefix of the tables to filter by.

        Returns:
            list: A list of table names.
        """
        try:
            # Initialize an empty list to hold table names
            table_names = []
            # Use the DynamoDB client to list tables
            # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb/client/list_tables.html
            response = self._db.meta.client.list_tables()
            _ = DEBUG and \
                log_debug(f'||| list_collections | response: {response}')
            # Extract table names from the response and add them to the list
            table_names.extend(
                [tn for tn in response.get('TableNames', [])
                 if not prefix or (prefix and tn.startswith(prefix))]
            )
            # Check if there are more tables to fetch
            while 'LastEvaluatedTableName' in response:
                # Fetch more tables starting from the last evaluated table name
                response = self._db.meta.client.list_tables(
                    ExclusiveStartTableName=response['LastEvaluatedTableName']
                )
                # Extract the next set of table names and add them to the list
                table_names.extend(
                    [tn for tn in response.get('TableNames', [])
                     if not prefix or (prefix and tn.startswith(prefix))]
                )
            # Optionally, filter table names if a collection_name is provided
            if collection_name:
                table_names = [name for name in table_names
                               if name == collection_name]
            # Return the list of table names
            _ = DEBUG and \
                log_debug(f'||| list_collections | table_names: {table_names}')
            return table_names
        except Exception as err:
            # Log the exception and return an empty list in case of an error
            log_debug(f"Error fetching table names: {str(err)}")
            return []

    def __getitem__(self, item_name):
        """
        Get the table object using the subscript operator.
        """
        table_name = f"{self._prefix}{item_name}"
        _ = DEBUG and \
            log_debug(f'||| __getitem__ | item_name: {item_name} |' +
                      f' table_name: {table_name}')
        return getattr(self, table_name)


class DynamodbServiceBuilder(DbAbstract):
    """
    Builder class for DynamodbService.
    """
    def __init__(self):
        self._instance = None

    def __call__(self, app_config: Config, **_ignored):
        if not self._instance:
            self._instance = DynamodbService(app_config)
        return self._instance


# ----------------------- Db General -----------------------


# Create a single instance of RequestHandler
# to be used throughout the module
request_handler = RequestHandler()


def set_db_request(request):
    """
    Set the request object for the database.
    This is used to get the current Request object.

    Args:
        request: The request object.

    Returns
        None.
    """
    request_handler.set_request(request)


def get_db_factory():
    """
    Get the database factory for the current database engine.

    Returns
        The database factory.

    Raises
        Exception: If the database engine is not supported.
    """
    settings = Config()
    factory = ObjectFactory()
    current_db_engine = settings.DB_ENGINE
    if DEBUG:
        log_debug(f'>>>--> current_db_engine = {current_db_engine}')
        # log_debug(f'>>>--> settings.DB_CONFIG = {settings.DB_CONFIG}')
    factory.register_builder('DYNAMO_DB', DynamodbServiceBuilder())
    factory.register_builder('MONGO_DB', MongodbServiceBuilder())
    return factory.create(current_db_engine, app_config=settings)


db_factory = LocalProxy(get_db_factory)


def get_db_engine():
    """
    Get the current database engine.
    """
    return db_factory.get_db()


db = LocalProxy(get_db_engine)


def test_connection():
    """
    Test database connection
    """
    return db_factory.test_connection()


# DB utilities


def verify_required_fields(fields: dict, required_fields: list,
                           error_code: str):
    """
    Verify if all the required fields are present in the fields dictionary.

    Args:
        fields (dict): The dictionary containing the fields.
        required_fields (list): The list of required fields.
        error_code (str): The error code to be returned.

    Returns:
        The resultset dictionary containing the error, error_message,
        and resultset. If there are no errors, the resultset will be empty
        and error = False. If there are errors, error = True and error_message
        will contain the missing fields.
    """
    resultset = dict(
        {
            'error': False,
            'error_message': '',
            'resultset': {}
        }
    )
    for element in required_fields:
        if element not in fields:
            # resultset['error_message'] = '{}{}{}'.format(
            #     resultset['error_message'],
            #     ', ' if resultset['error_message'] != '' else '', element
            # )
            resultset['error_message'] = f"{resultset['error_message']}" + \
                f"{', ' if resultset['error_message'] != '' else ''}{element}"
    if resultset['error_message']:
        resultset['error_message'] = 'Missing mandatory elements:' + \
            f" {resultset['error_message']} {error_code}."
        resultset['error'] = True
    return resultset


def get_order_direction(direction: str):
    """
    Get the order direction.

    Args:
        direction (str): The direction to be returned.

    Returns:
        The order direction with the MongoDb constants.
    """
    return pymongo.ASCENDING if direction == "asc" else pymongo.DESCENDING



"""
Configuration manager
"""
# C0103 | Disable "name doesn't conform to naming rules..." (snake_case)
# pylint: disable=C0103
# R0902 | Disable "too-many-instance-attributes"
# pylint: disable=R0902
# R0903 | Disable "too-few-public-methods"
# pylint: disable=R0903
# R0915 | Disable "too-many-statements "
# pylint: disable=R0915
# W0105 | Disable "pointless-string-statement" (for """ comments)
# pylint: disable=W0105
# C0301: | Disable "line-too-long"
# pylint: disable=C0301

from typing import Union, Any
import os
import json
import logging
import datetime


from genericsuite.config.config_secrets import get_secrets_from_iaas


def get_default_resultset() -> dict:
    """Returns an standard base resultset, to be used in the building
       of responses to the outside world
    """
    resultset = {
        'error': False,
        'error_message': None,
        'totalPages': None,
        'resultset': {}
    }
    return resultset


def formatted_log_message(message: str) -> str:
    """ Returns a formatted message with database name and date/time """
    return f"[{os.environ.get('APP_DB_NAME', 'APP_DB_NAME not set')}]" + \
        f" {datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}" + \
        f" | {message}"


def get_config_logger() -> logging.Logger:
    """
    Get the logger object.
    """
    return logging.getLogger(os.environ.get('APP_NAME'))


def config_log_error(message: str) -> None:
    """
    Log a message to the console.
    """
    get_config_logger().error("%s", formatted_log_message(message))


def config_log_debug(message: str) -> None:
    """
    Log a message to the console.
    """
    get_config_logger().debug("%s", formatted_log_message(message))


def config_log_info(message: str) -> None:
    """
    Log a message to the console.
    """
    get_config_logger().debug("%s", formatted_log_message(message))


def text_to_dict(text: str) -> Union[dict, None]:
    """
    Convert a text string to a dictionary.
    """
    result = None
    try:
        result = json.loads(text)
    except json.JSONDecodeError as e:
        config_log_error(f'ERROR [C-E010] | converting text to dict: {e}')
    return result


def is_local_service() -> bool:
    return os.environ.get('AWS_SAM_LOCAL') == 'true' or \
        os.environ.get('GS_LOCAL_ENVIR') == 'true'


class Config():
    """ Configuration class, to have the most used App variables """
    def __init__(self, app_context: Any = None) -> None:

        # Set the local app_context to eventually get values from
        # Database (any other place than the App initialization)
        self.app_context = app_context

        # Get secrets and set environment variables
        params = get_secrets_from_iaas(get_default_resultset,
                                       get_config_logger())
        if params["error"]:
            error_msg = 'CNFG-1) ERROR: Config.__init__() |' + \
                        f' Getting Secrets | params: {params}'
            raise Exception(error_msg)

        # ............................

        # IMPORTANT: these parameters values must be always retrieved
        # from environment variables

        # Database configuration

        if is_local_service():
            # Handles the \@ issue in environment variables values when runs
            # by "sam local start-api"
            os.environ['APP_DB_URI'] = \
                os.environ['APP_DB_URI'].replace('\\@', '@')
            os.environ['APP_SUPERADMIN_EMAIL'] = \
                os.environ['APP_SUPERADMIN_EMAIL'].replace('\\@', '@')

        self.DB_CONFIG = {
            'mongodb_uri': os.environ['APP_DB_URI'],
            'mongodb_db_name': os.environ['APP_DB_NAME'],
            'dynamdb_prefix': os.environ.get('DYNAMDB_PREFIX', ''),
        }
        # DB_ENGINE = 'MONGO_DB'
        # DB_ENGINE = 'DYNAMO_DB'
        self.DB_ENGINE = os.environ['APP_DB_ENGINE']

        # App general configuration

        self.DEBUG = self.get_env('APP_DEBUG', '0') == '1'

        self.APP_NAME = os.environ['APP_NAME']
        self.APP_VERSION = os.environ.get('APP_VERSION', 'N/A')
        self.STAGE = os.environ.get('APP_STAGE')
        self.SECRET_KEY = os.environ.get('SECRET_KEY', str(os.urandom(16)))

        # App specific configuration

        self.APP_SECRET_KEY = os.environ['APP_SECRET_KEY']
        self.APP_SUPERADMIN_EMAIL = \
            os.environ['APP_SUPERADMIN_EMAIL']

        self.APP_HOST_NAME = os.environ['APP_HOST_NAME']
        self.STORAGE_URL_SEED = os.environ['STORAGE_URL_SEED']

        self.GIT_SUBMODULE_LOCAL_PATH = os.environ['GIT_SUBMODULE_LOCAL_PATH']

        self.TEMP_DIR = os.environ.get('TEMP_DIR', '/tmp')

        # ............................

        # Auth parameters

        self.CORS_ORIGIN = self.get_env('APP_CORS_ORIGIN', '*')
        self.HEADER_TOKEN_ENTRY_NAME = self.get_env(
            'HEADER_TOKEN_ENTRY_NAME',
            'Authorization'  # 'x-access-tokens'
        )

        # Languages

        self.DEFAULT_LANG = self.get_env('DEFAULT_LANG', 'en')

    def get_env(self, var_name: str, def_value: Any = None) -> Any:
        """
        Get value of a config variable. If it's in the app_context,
        get from there, if not, get from os.environ.
        """
        result = os.environ.get(var_name, def_value)
        if self.app_context:
            result = self.app_context.get_env_var(
                var_name=var_name,
                def_value=result,
            )
        return result
        # return getattr(self, var_name)

    def debug_vars(self) -> str:
        """
        Show all defined config variables.
        """
        return (
            'Config.debug_vars:\n\n' +
            f'DEBUG = {self.DEBUG}\n' +
            f'SECRET_KEY = {self.SECRET_KEY}\n' +
            f'DB_CONFIG = {self.DB_CONFIG}\n' +
            f'DB_ENGINE = {self.DB_ENGINE}\n' +
            f'APP_SECRET_KEY = {self.APP_SECRET_KEY}\n' +
            f'APP_SUPERADMIN_EMAIL = {self.APP_SUPERADMIN_EMAIL}\n' +
            f'CORS_ORIGIN = {self.CORS_ORIGIN}\n' +
            f'HEADER_TOKEN_ENTRY_NAME = {self.HEADER_TOKEN_ENTRY_NAME}\n' +
            f'STAGE = {self.STAGE}\n' +
            '\n'
        )



"""
Menu options access
"""
from typing import Optional

# from chalice.app import Response
# from genericsuite.util.blueprint_one import BlueprintOne
from genericsuite.util.framework_abs_layer import Response, BlueprintOne

from genericsuite.util.jwt import (
    request_authentication,
    AuthorizedRequest
)

from genericsuite.models.menu_options.menu_options import (
    menu_options_get as menu_options_get_model,
    menu_options_element as menu_options_element_model,
)

bp = BlueprintOne(__name__)


@bp.route(
    '/',
    methods=['GET'],
    authorizor=request_authentication(),
)
def menu_options_get(
    request: AuthorizedRequest,
    other_params: Optional[dict] = None
) -> Response:
    """ Get authorized menu options """
    # return menu_options_get_model(request, other_params)
    return menu_options_get_model(request, bp, other_params)


@bp.route(
    '/element',
    methods=['POST'],
    authorizor=request_authentication(),
)
def menu_options_element(
    request: AuthorizedRequest,
    other_params: Optional[dict] = None
) -> Response:
    """ Get menu element configuration """
    # return menu_options_element_model(request, other_params)
    return menu_options_element_model(request, bp, other_params)



"""
Constant tables (dict)
"""
from typing import Any
from genericsuite.util.config_dbdef_helpers import get_json_def_both


def get_all_constants() -> dict:
    """
    Get all constants from the json files: general_constants.json
    and app_constants.json

    Returns:
        dict: a dict with all constants attributes in the json files.
    """
    result = get_json_def_both("general_constants")
    result.update(get_json_def_both("app_constants"))
    return result


def get_constant(const_name: str, entry_name: str = None,
                 def_value: Any = None) -> Any:
    """
    Get a constant value from the constants attribute const_name.
    If the entry_name is not found in the constants attribute,
    returns the def_value.
    If the entry_name is None, returns all the constants attribute content.

    Args:
        const_name (str): Constant's attribute name
        entry_name (str): Entry name to be found in the constant's attribute
        name.
        def_value (Any): default value when the entry name is not found.
        Defaults to None.

    Returns:
        Any: the constant value from the constants attribute or the default
        value if it doesn't exist. If the entry_name is None, returns all
        the constants attribute content.
    """
    if not entry_name:
        return constants[const_name]
    return constants[const_name].get(entry_name, def_value)


constants = get_all_constants()



"""
FastAPI abstraction layer
"""
# from typing import Optional, Union, Dict, Any
from typing import Optional, Dict, Any, Tuple, Callable, List, Union

import os
import importlib
import json

import jwt
from pydantic import BaseModel

from fastapi import Request as FastAPIRequest
from fastapi import Response as FastAPIResponse

from fastapi import APIRouter, Depends, HTTPException
from fastapi.security import OAuth2PasswordBearer

# from genericsuite.fastapilib.util.blueprint_one import (
#     BlueprintOne as FaBlueprintOne
# )

from genericsuite.config.config import (
    Config,
    config_log_debug as log_debug,
    config_log_error as log_error,
)

DEBUG = False

oauth2_scheme = OAuth2PasswordBearer(tokenUrl="token")

FRAMEWORK_LOADED = False
FRAMEWORK = os.environ.get('CURRENT_FRAMEWORK', '').lower()
if FRAMEWORK == 'fastapi':
    try:
        framework_module = importlib.import_module(FRAMEWORK)
        FRAMEWORK_LOADED = True
        if DEBUG:
            print(f'FastAPI abstraction | framework_module: {framework_module}')


        class FrameworkClass(framework_module.FastAPI):
            """
            Framkework class cloned from the selected framework super class.
            """


        class Request(BaseModel):
            """
            Request class cloned from the selected Request framework super class.
            This class is the one to be imported by the project modules
            """
            method: Optional[str] = "GET"
            query_params: Optional[dict] = {}
            json_body: Optional[dict] = {}
            headers: Optional[dict] = {}
            event_dict: Optional[Dict[str, Any]] = {}
            lambda_context: Optional[Any] = None

            def to_dict(self):
                """
                Returns the request data as a dictionary.
                """
                return {
                    "method": self.method,
                    "query_params": self.query_params,
                    "json_body": self.json_body,
                    "headers": self.headers,
                }

            def to_original_event(self) -> Union[Dict[str, Any], None]:
                """
                Returns the original event dictionary.
                """
                return self.event_dict


        class Response(FastAPIResponse):
            """
            Response class cloned from the selected Response framework super class.
            This class is the one to be imported by the project modules
            """
            body: Union[str, dict]
            status_code: Optional[int] = 200
            headers: Optional[dict] = {}

            def __init__(
                self,
                body: Union[str, dict],
                status_code: Optional[int] = 200,
                headers: Optional[dict] = None
            ):
                """
                Initializes the Response object.
                """
                if isinstance(body, dict):
                    body = json.dumps(body)

                headers = headers if headers else {}
                if 'Content-Type' not in headers:
                    headers['Content-Type'] = 'application/json'
                if 'Access-Control-Allow-Origin' not in headers:
                    headers["Access-Control-Allow-Origin"] = os.environ.get('APP_CORS_ORIGIN', '*')
                if 'Access-Control-Allow-Methods' not in headers:
                    headers["Access-Control-Allow-Methods"] = "GET, POST, PUT, DELETE, OPTIONS"
                if 'Access-Control-Allow-Headers' not in headers:
                    headers["Access-Control-Allow-Headers"] = "Content-Type, Authorization"

                if DEBUG:
                    print('FastAPI abstraction' +
                          f'\n| body: {body}' +
                          f'\n| status_code: {status_code}' +
                          f'\n| headers: {headers}')
                super().__init__(
                    content=body,
                    status_code=status_code,
                    headers=headers
                )


        class Blueprint(framework_module.APIRouter):
            """
            Blueprint class cloned from the selected Blueprint framework super class.
            This class is the one to be imported by the project modules
            """


        # FastAPI BlueprintOne implementation

        # JWT

        class AuthTokenPayload(BaseModel):
            """
            Represents the user's payload structure of the JWT token.
            """
            public_id: str


        class AuthorizedRequest(Request):
            """
            Represents the AuthorizedRequest payload structure of the JWT token,
            containing the authenticated user's data.
            """
            user: AuthTokenPayload


        def credentials_exception(
            error_message: Optional[str] = None,
            error_code: Optional[int] = 401,  # Unauthorized
            headers: Optional[dict] = None
        ) -> HTTPException:
            """
            Returns a standard error response
            """
            if not headers:
                headers = {"WWW-Authenticate": "Bearer"}
            if not error_message:
                error_message = 'Could not validate credentials'
            return HTTPException(
                status_code=error_code,
                detail=error_message,
                headers=headers,
            )


        def get_general_authorized_request(
            request: Request
        ) -> Union[AuthorizedRequest, HTTPException]:
            """
            Get the authorized request from the request object.

            Args:
                request (Request): The request object.

            Returns:
                AuthorizedRequest: The authorized request.
            """
            settings = Config()
            if settings.HEADER_TOKEN_ENTRY_NAME not in request.headers:
                # return standard_error_return('A valid token is missing')
                return credentials_exception('A valid token is missing')
            try:
                token_raw = request.headers[settings.HEADER_TOKEN_ENTRY_NAME]
                jwt_token = token_raw.replace('Bearer ', '')
                if DEBUG:
                    log_debug('||| FA REQUEST_AUTHENTICATION' +
                        '\n | HEADER_TOKEN_ENTRY_NAME: ' +
                        f'{settings.HEADER_TOKEN_ENTRY_NAME}' +
                        f'\n | token_raw: {token_raw}' +
                        f'\n | jwt_token: {jwt_token}' +
                        # f'\n | settings.APP_SECRET_KEY: {settings.APP_SECRET_KEY}' +
                        '\n')
                jws_token_data = jwt.decode(
                    jwt_token,
                    settings.APP_SECRET_KEY,
                    algorithms="HS256",
                )
                if DEBUG:
                    log_debug('||| FA REQUEST_AUTHENTICATION' +
                        f' | jws_token_data = {jws_token_data}')

                authorized_request = AuthorizedRequest(
                    # type: ignore[attr-defined]
                    event_dict=request.to_original_event(),
                    # type: ignore[attr-defined]
                    lambda_context=request.lambda_context,
                    # Authentication token data
                    user=jws_token_data,
                )

                if DEBUG:
                    log_debug('||| FA REQUEST_AUTHENTICATION' +
                        f' | authorized_request = {authorized_request}')
            except Exception as err:
                log_error('REQUEST_AUTHENTICATION' +
                    f' | Exception = {str(err)}')
                # return standard_error_return('Token is invalid')
                return credentials_exception('Token is invalid')
            return authorized_request


        # Dependencies

        def build_request(
            method: Optional[str] = None,
            query_params: Optional[dict] = None,
            json_body: Optional[dict] = None,
            headers: Optional[dict] = None,
            preserve_nones: bool = False,
        ) -> Union[Request, AuthorizedRequest]:
            """
            Builds the request query parameters from query_received and cleans it,
            leaving only the ones that are not None

            Args:
                method (str): HTTP method.
                query_params (dict): URL query parameters.
                json_body (dict): Request JSON body.
                headers (dict): Request headers.

            Returns:
                dict: The cleaned request query parameters.
            """
            query_params_reduced = {}
            if query_params:
                # Reduce query_params leaving only the not None items
                query_params_reduced = {k: v for k, v in query_params.items()
                                        if v is not None or preserve_nones}
            headers_reduced = headers if headers else {}
            if "token" in headers_reduced:
                headers_reduced["Authorization"] = f"Bearer {headers_reduced['token']}"
                del headers_reduced["token"]
            if "current_user" in headers_reduced:
                new_request = AuthorizedRequest(
                    method=method if method else "get",
                    query_params=query_params_reduced,
                    json_body=json_body if json_body else {},
                    headers=headers_reduced,
                    user=headers_reduced["current_user"])
            else:
                new_request = Request(
                    method=method if method else "get",
                    query_params=query_params_reduced,
                    json_body=json_body if json_body else {},
                    headers=headers_reduced)
            return new_request


        async def get_current_user(token: str = Depends(oauth2_scheme)):
            """
            Verifies the JWT token and returns the current user.
            """
            request = build_request(headers={"token": token})
            auth_request = get_general_authorized_request(request)
            if not isinstance(auth_request, AuthorizedRequest):
                raise auth_request
            return getattr(auth_request, 'user', None)


        def get_default_fa_request(
            method: Optional[str] = None,
            current_user: Optional[AuthTokenPayload] = None,
            json_body: Optional[dict] = None,
            headers: Optional[dict] = None,
            query_params: Optional[dict] = None,
            preserve_nones: bool = False,
        ) -> Union[Request, AuthorizedRequest]:
            """
            Builds the default FA (FastAPI) Authentication request object.
            """
            params = {
                "method": method or 'get',
                "json_body": json_body if json_body else {},
                "headers": headers if headers else {},
                "query_params": query_params if query_params else {},
                "preserve_nones": preserve_nones,
            }
            if current_user:
                params["headers"]["current_user"] = current_user
            request = build_request(**params)
            return request


        class BlueprintOne(APIRouter):
            """
            Class to register a new route with optional schema validation and authorization.
            """
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.request = None
                self.other_params = None

            def get_current_app(self):
                """
                Get the current App object. It must be called inside a router function.
                """
                return self

            def get_current_request(self):
                """
                Get the current App object. It must be called inside a router function.
                """
                return self.request

            def get_other_params(self):
                """
                Get the other_params object.
                """
                return self.other_params

            def api_route(
                self,
                path: str,
                methods: Union[str, List[str]],
                requires_auth: Optional[bool] = True,
                preserve_nones: Optional[bool] = False,
                other_params: Optional[dict] = None,
                **kwargs
            ):
                def decorator(func: Callable):
                    async def endpoint_wrapper(request: Request, *args, **kwargs):
                        current_user = None
                        if requires_auth:
                            current_user = await get_current_user()

                        json_body = await request.json() \
                            if request.method in ["POST", "PUT", "PATCH"] \
                            else None
                        headers = dict(request.headers)
                        query_params = dict(request.query_params)

                        request_obj = get_default_fa_request(
                            method=request.method,
                            current_user=current_user,
                            json_body=json_body,
                            headers=headers,
                            query_params=query_params,
                            preserve_nones=preserve_nones,
                        )
                        self.request = request_obj
                        self.other_params = other_params or {}

                        return await func(request, *args, **kwargs)

                    self.add_api_route(
                        path=path,
                        endpoint=endpoint_wrapper,
                        methods=methods,
                        include_in_schema=True,
                        **kwargs
                    )
                    return endpoint_wrapper

                return decorator

            def get(
                self,
                path: str,
                requires_auth: Optional[bool] = True,
                preserve_nones: Optional[bool] = False,
                other_params: Optional[dict] = None,
                **kwargs
            ):
                return self.api_route(
                    path,
                    methods=["GET"],
                    requires_auth=requires_auth,
                    preserve_nones=preserve_nones,
                    other_params=other_params,
                    **kwargs
                )

            def post(
                self,
                path: str,
                requires_auth: Optional[bool] = True,
                preserve_nones: Optional[bool] = False,
                other_params: Optional[dict] = None,
                **kwargs
            ):
                return self.api_route(
                    path,
                    methods=["POST"],
                    requires_auth=requires_auth,
                    preserve_nones=preserve_nones,
                    other_params=other_params,
                    **kwargs
                )

            def put(
                self,
                path: str,
                requires_auth: Optional[bool] = True,
                preserve_nones: Optional[bool] = False,
                other_params: Optional[dict] = None,
                **kwargs
            ):
                return self.api_route(
                    path,
                    methods=["PUT"],
                    requires_auth=requires_auth,
                    preserve_nones=preserve_nones,
                    other_params=other_params,
                    **kwargs
                )

            def delete(
                self,
                path: str,
                requires_auth: Optional[bool] = True,
                preserve_nones: Optional[bool] = False,
                other_params: Optional[dict] = None,
                **kwargs
            ):
                return self.api_route(
                    path,
                    methods=["DELETE"],
                    requires_auth=requires_auth,
                    preserve_nones=preserve_nones,
                    other_params=other_params,
                    **kwargs
                )


    except ImportError as frm_err:
        raise ImportError(f"Unable to import '{FRAMEWORK}': {frm_err}") from None



"""
Generic Database middleware to facilitate calls to GenericDbHelper
"""
from typing import List, Union, Optional

from genericsuite.util.app_logger import log_debug, log_error
from genericsuite.util.generic_db_helpers import GenericDbHelper
from genericsuite.util.app_context import AppContext

DEBUG = False


# Database generic operations


def fetch_all_from_db(
    app_context: AppContext,
    json_file: str,
    like_query_params: Optional[dict] = None,
    combinator: Optional[str] = None,
    order_param: Optional[str] = None,
) -> dict:
    """
    Fetches all items from the database table using the GenericDbHelper class.

    Args:
        json_file (str): The JSON file containing the database.
        like_query_params (dict): The query parameters to filter
        the items as SQL LIKE.
        combinator (str): condiition combinator for the filter.
        It could be $and, $or, $not. Defaults to $and

    Returns:
        List[dict]: The fetched items resultset.
    """
    dbo = GenericDbHelper(
        json_file=json_file,
        request=app_context.get_request(),
        blueprint=app_context.get_blueprint(),
    )
    result = dbo.fetch_list(
        skip=0,
        limit=0,
        like_query_params=like_query_params,
        combinator=combinator,
        order_param=order_param,
    )
    if result['error']:
        log_error(
            f'AI_FAFD-E1) ERROR: Fetch all from {json_file}' +
            ' | result: {result}')
    elif DEBUG:
        log_debug(f'AI_FAFD-1) Fetch all from {json_file} | result: {result}')
    return result


def get_item_from_db(
    app_context: AppContext,
    json_file: str,
    entry_name: str,
    entry_value: Union[str, int, float],
    user_id: str = None,
) -> dict:
    """
    Fetches an item from the database by an entry name (field/column).

    Args:
        json_file (str): The JSON file containing the database.
        entry_name (str): The name of the entry to fetch.
        entry_value (Union[str, int]): The value of the entry to fetch.

    Returns:
        dict: The fetched item resultset.
    """
    dbo = GenericDbHelper(
        json_file=json_file,
        request=app_context.get_request(),
        blueprint=app_context.get_blueprint(),
    )
    filters = {}
    if user_id:
        filters = {"user_id": user_id}
    get_by_pk = entry_name in ["id", "_id"]
    if get_by_pk:
        result = dbo.fetch_row_raw(entry_value)
    else:
        result = dbo.fetch_row_by_entryname_raw(
            entry_name=entry_name,
            entry_value=entry_value,
            filters=filters
        )
    if result['error']:
        log_error(
            f'AG_IFD-E1) ERROR get_item_from_db {json_file} |' +
            f' get_by_pk: {get_by_pk} | result: {result}')
    elif DEBUG:
        log_debug(
            f'AG_IFD-1) get_item_from_db {json_file} |' +
            f' get_by_pk: {get_by_pk} | result: {result}')
    return result


def modify_item_in_db(
    app_context: AppContext,
    json_file: str,
    data: dict
) -> dict:
    """Modify an Item (row) in the database

    Args:
        json_file (str): json file name with the table definition.
        entry_name (str): The name of the entry to modify.
        entry_value (Union[str, int, float]): The value of the entry to modify.
        data (dict): data to be updated

    Returns:
        dict: operation result as a resultset. See: update_row()
    """
    dbo = GenericDbHelper(
        json_file=json_file,
        request=app_context.get_request(),
        blueprint=app_context.get_blueprint(),
    )
    result = dbo.update_row(data)
    if result['error']:
        log_error(
            'AI_MIFD-E1) ERROR modify_item_in_db {json_file}' +
            f' | result: {result}')
    elif DEBUG:
        log_debug(f'AI_MIFD-1) modify_item_in_db {json_file}' +
                  f' | result: {result}')
    return result


def add_item_to_db(
    app_context: AppContext,
    json_file: str,
    data: dict,
    filters: dict = None,
) -> dict:
    """Add an Item (row) to the database

    Args:
        json_file (str): json file name with the table definition.
        data (dict): data to be registered

    Returns:
        dict: operation result as a resultset. See: create_row()
    """
    filters = {} if not filters else filters
    dbo = GenericDbHelper(
        json_file=json_file,
        request=app_context.get_request(),
        blueprint=app_context.get_blueprint(),
    )
    result = dbo.create_row(data, filters)
    if result['error']:
        log_error(
            f'AI_AITD-E1) ERROR add_item_to_db {json_file} | result: {result}')
    elif DEBUG:
        log_debug(f'AI_AITD-1) add_item_to_db {json_file} | result: {result}')
    return result


def fetch_all_from_db_array(
    app_context: AppContext,
    json_file: str,
    row_id: str,
    like_query_params: Optional[dict] = None,
    combinator: Optional[str] = None,
    order_param: str = None,
# ) -> List[dict]:
) -> dict:
    """
    Fetches all items in a row's array from the database table
    using the GenericDbHelper class.

    Args:
        json_file (str): The JSON file containing the database.
        row_id (str): parent row ID.
        like_query_params (dict): The query parameters to filter
        the items as SQL LIKE.
        combinator (str): condiition combinator for the filter.
        It could be $and, $or, $not. Defaults to $and

    Returns:
        List[dict]: The fetched items resultset.
    """
    dbo = GenericDbHelper(
        json_file=json_file,
        request=app_context.get_request(),
        blueprint=app_context.get_blueprint(),
    )
    result = dbo.fetch_array_rows(
        row_id=row_id,
        skip=0,
        limit=0,
        like_query_params=like_query_params,
        combinator=combinator,
        order_param=order_param,
    )
    if result['error']:
        log_error(
            f'AI_FAFDA-E1) ERROR Fetch all (array) from {json_file} |' +
            f' result: {result}')
    elif DEBUG:
        log_debug(f'AI_FAFDA-1) Fetch all (array) from {json_file} |' +
                  f' result: {result}')
    return result



"""
Menu options access
"""
# from fastapi import APIRouter, Depends, Body
from fastapi import Depends, Body, Request as FaRequest
from pydantic import BaseModel

from genericsuite.util.framework_abs_layer import Response
from genericsuite.fastapilib.util.blueprint_one import BlueprintOne
from genericsuite.fastapilib.util.dependencies import (
    get_current_user,
    get_default_fa_request,
)
from genericsuite.models.menu_options.menu_options import (
    menu_options_get as menu_options_get_model,
    menu_options_element as menu_options_element_model,
)


class MenuElementRequest(BaseModel):
    """ Menu element request """
    element: str


# router = APIRouter()
router = BlueprintOne()


@router.get('')
async def menu_options_get(
    request: FaRequest,
    current_user: str = Depends(get_current_user),
) -> Response:
    """ Get authorized menu options """
    gs_request, other_params = get_default_fa_request(current_user)
    router.set_current_request(request, gs_request)
    return menu_options_get_model(request=gs_request, blueprint=router,
        other_params=other_params)


@router.post(
    '/element',
    tags=['element'],
)
async def menu_options_element(
    request: FaRequest,
    current_user: str = Depends(get_current_user),
    json_body: MenuElementRequest = Body(...),
) -> Response:
    """ Get menu element configuration """
    gs_request, other_params = get_default_fa_request(current_user,
        json_body=json_body.model_dump())
    router.set_current_request(request, gs_request)
    return menu_options_element_model(request=gs_request, blueprint=router,
        other_params=other_params)
